"""

Revision ID: 08ac1116e055
Revises: 11d35a452b40
Create Date: 2025-06-27 21:22:44.798799

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision: str = "08ac1116e055"
down_revision: Union[str, None] = "11d35a452b40"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("cdlmorningstar", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("cdl3linestrike", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("cdl3whitesoldiers", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("cdlabandonedbaby", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("cdltasukigap", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("cdlpiercing", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("cdlengulfing", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("rate_of_change_1", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("rate_of_change_7_4", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("rate_of_change_7_12", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("rate_of_change_30", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("golden_cross", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("death_cross", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("mfi_oversold", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("mfi_overbought", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("stoch_oversold", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("stoch_overbought", sa.Date(), nullable=True))
        batch_op.add_column(
            sa.Column("rsi_bullish_crossover", sa.Date(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("rsi_bearish_crossover", sa.Date(), nullable=True)
        )
        batch_op.add_column(sa.Column("rsi_oversold", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("rsi_overbought", sa.Date(), nullable=True))
        batch_op.add_column(
            sa.Column("macd_12_26_9_bullish_crossover", sa.Date(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("macd_12_26_9_bearish_crossover", sa.Date(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("macd_12_26_9_zero_line_cross_up", sa.Date(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("macd_12_26_9_zero_line_cross_down", sa.Date(), nullable=True)
        )
        batch_op.add_column(sa.Column("adx_14_long", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("adx_14_short", sa.Date(), nullable=True))
        batch_op.create_index("ix_analysis_adx_14_long", ["adx_14_long"], unique=False)
        batch_op.create_index(
            "ix_analysis_adx_14_short", ["adx_14_short"], unique=False
        )
        batch_op.create_index("ix_analysis_base_symbol", ["base_symbol"], unique=False)
        batch_op.create_index("ix_analysis_book_value", ["book_value"], unique=False)
        batch_op.create_index(
            "ix_analysis_cdl3linestrike", ["cdl3linestrike"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_cdl3whitesoldiers", ["cdl3whitesoldiers"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_cdlabandonedbaby", ["cdlabandonedbaby"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_cdlengulfing", ["cdlengulfing"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_cdlmorningstar", ["cdlmorningstar"], unique=False
        )
        batch_op.create_index("ix_analysis_cdlpiercing", ["cdlpiercing"], unique=False)
        batch_op.create_index(
            "ix_analysis_cdltasukigap", ["cdltasukigap"], unique=False
        )
        batch_op.create_index("ix_analysis_country", ["country"], unique=False)
        batch_op.create_index("ix_analysis_created_at", ["created_at"], unique=False)
        batch_op.create_index("ix_analysis_currency", ["currency"], unique=False)
        batch_op.create_index("ix_analysis_date", ["date"], unique=False)
        batch_op.create_index("ix_analysis_death_cross", ["death_cross"], unique=False)
        batch_op.create_index(
            "ix_analysis_dividend_rate", ["dividend_rate"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_dividend_yield", ["dividend_yield"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_earning_per_share", ["earning_per_share"], unique=False
        )
        batch_op.create_index("ix_analysis_exchange", ["exchange"], unique=False)
        batch_op.create_index(
            "ix_analysis_forward_earnings_per_share",
            ["forward_earnings_per_share"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_golden_cross", ["golden_cross"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_gross_margins", ["gross_margins"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_growing_basic_eps", ["growing_basic_eps"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_growing_diluted_eps", ["growing_diluted_eps"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_growing_net_income", ["growing_net_income"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_growing_operating_cash_flow",
            ["growing_operating_cash_flow"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_growing_operating_income",
            ["growing_operating_income"],
            unique=False,
        )
        batch_op.create_index("ix_analysis_industry", ["industry"], unique=False)
        batch_op.create_index(
            "ix_analysis_industry_group", ["industry_group"], unique=False
        )
        batch_op.create_index("ix_analysis_isin", ["isin"], unique=False)
        batch_op.create_index("ix_analysis_last_price", ["last_price"], unique=False)
        batch_op.create_index(
            "ix_analysis_macd_12_26_9_bearish_crossover",
            ["macd_12_26_9_bearish_crossover"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_macd_12_26_9_bullish_crossover",
            ["macd_12_26_9_bullish_crossover"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_macd_12_26_9_zero_line_cross_down",
            ["macd_12_26_9_zero_line_cross_down"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_macd_12_26_9_zero_line_cross_up",
            ["macd_12_26_9_zero_line_cross_up"],
            unique=False,
        )
        batch_op.create_index("ix_analysis_market", ["market"], unique=False)
        batch_op.create_index(
            "ix_analysis_market_capitalization", ["market_capitalization"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_max_capex_ratio", ["max_capex_ratio"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_max_dividend_payout_ratio",
            ["max_dividend_payout_ratio"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_mean_capex_ratio", ["mean_capex_ratio"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_mean_dividend_payout_ratio",
            ["mean_dividend_payout_ratio"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_mfi_overbought", ["mfi_overbought"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_mfi_oversold", ["mfi_oversold"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_min_capex_ratio", ["min_capex_ratio"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_min_dividend_payout_ratio",
            ["min_dividend_payout_ratio"],
            unique=False,
        )
        batch_op.create_index("ix_analysis_modifier", ["modifier"], unique=False)
        batch_op.create_index("ix_analysis_name", ["name"], unique=False)
        batch_op.create_index(
            "ix_analysis_operating_cash_flow_is_higher_than_net_income",
            ["operating_cash_flow_is_higher_than_net_income"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_operating_margins", ["operating_margins"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_positive_basic_eps", ["positive_basic_eps"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_positive_debt_to_equity",
            ["positive_debt_to_equity"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_positive_diluted_eps", ["positive_diluted_eps"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_positive_free_cash_flow",
            ["positive_free_cash_flow"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_positive_net_income", ["positive_net_income"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_positive_operating_income",
            ["positive_operating_income"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_positive_return_on_assets",
            ["positive_return_on_assets"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_positive_return_on_equity",
            ["positive_return_on_equity"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_price_per_earning_ratio",
            ["price_per_earning_ratio"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_price_to_book", ["price_to_book"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_quarterly_earning_per_share",
            ["quarterly_earning_per_share"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_growing_basic_eps",
            ["quarterly_growing_basic_eps"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_growing_diluted_eps",
            ["quarterly_growing_diluted_eps"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_growing_net_income",
            ["quarterly_growing_net_income"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_growing_operating_cash_flow",
            ["quarterly_growing_operating_cash_flow"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_growing_operating_income",
            ["quarterly_growing_operating_income"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_max_capex_ratio",
            ["quarterly_max_capex_ratio"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_max_dividend_payout_ratio",
            ["quarterly_max_dividend_payout_ratio"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_mean_capex_ratio",
            ["quarterly_mean_capex_ratio"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_mean_dividend_payout_ratio",
            ["quarterly_mean_dividend_payout_ratio"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_min_capex_ratio",
            ["quarterly_min_capex_ratio"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_min_dividend_payout_ratio",
            ["quarterly_min_dividend_payout_ratio"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_operating_cash_flow_is_higher_than_net_income",
            ["quarterly_operating_cash_flow_is_higher_than_net_income"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_positive_basic_eps",
            ["quarterly_positive_basic_eps"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_positive_debt_to_equity",
            ["quarterly_positive_debt_to_equity"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_positive_diluted_eps",
            ["quarterly_positive_diluted_eps"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_positive_free_cash_flow",
            ["quarterly_positive_free_cash_flow"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_positive_net_income",
            ["quarterly_positive_net_income"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_positive_operating_income",
            ["quarterly_positive_operating_income"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_positive_return_on_assets",
            ["quarterly_positive_return_on_assets"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_quarterly_positive_return_on_equity",
            ["quarterly_positive_return_on_equity"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_rate_of_change_1", ["rate_of_change_1"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_rate_of_change_30", ["rate_of_change_30"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_rate_of_change_7_12", ["rate_of_change_7_12"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_rate_of_change_7_4", ["rate_of_change_7_4"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_return_on_equity", ["return_on_equity"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_revenue_growth", ["revenue_growth"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_rsi_bearish_crossover", ["rsi_bearish_crossover"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_rsi_bullish_crossover", ["rsi_bullish_crossover"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_rsi_overbought", ["rsi_overbought"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_rsi_oversold", ["rsi_oversold"], unique=False
        )
        batch_op.create_index("ix_analysis_sector", ["sector"], unique=False)
        batch_op.create_index(
            "ix_analysis_stoch_overbought", ["stoch_overbought"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_stoch_oversold", ["stoch_oversold"], unique=False
        )
        batch_op.create_index("ix_analysis_summary", ["summary"], unique=False)
        batch_op.create_index(
            "ix_analysis_trailing_earnings_per_share",
            ["trailing_earnings_per_share"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_trailing_price_to_earnings",
            ["trailing_price_to_earnings"],
            unique=False,
        )
        batch_op.create_index("ix_analysis_website", ["website"], unique=False)
        batch_op.drop_column("last_year_growth")
        batch_op.drop_column("year_to_date_max_growth")
        batch_op.drop_column("last_month_max_growth")
        batch_op.drop_column("last_week_max_growth")
        batch_op.drop_column("star_wow")
        batch_op.drop_column("last_dmp")
        batch_op.drop_column("year_to_date_growth")
        batch_op.drop_column("last_week_growth")
        batch_op.drop_column("rsi_last_value")
        batch_op.drop_column("ma_50_200_buy_date")
        batch_op.drop_column("slope_14")
        batch_op.drop_column("slope_7")
        batch_op.drop_column("last_52_weeks_growth")
        batch_op.drop_column("slope_60")
        batch_op.drop_column("last_dmn")
        batch_op.drop_column("last_price_date")
        batch_op.drop_column("slope_30")
        batch_op.drop_column("macd_12_26_9_buy")
        batch_op.drop_column("last_year_max_growth")
        batch_op.drop_column("macd_12_26_9_buy_date")
        batch_op.drop_column("last_adx")
        batch_op.drop_column("last_month_growth")
        batch_op.drop_column("star_mom")
        batch_op.drop_column("star_yoy")

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("star_yoy", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("star_mom", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("last_month_growth", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("last_adx", sa.FLOAT(), nullable=True))
        batch_op.add_column(
            sa.Column("macd_12_26_9_buy_date", sa.DATE(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("last_year_max_growth", sa.FLOAT(), nullable=True)
        )
        batch_op.add_column(sa.Column("macd_12_26_9_buy", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("slope_30", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("last_price_date", sa.DATE(), nullable=True))
        batch_op.add_column(sa.Column("last_dmn", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("slope_60", sa.FLOAT(), nullable=True))
        batch_op.add_column(
            sa.Column("last_52_weeks_growth", sa.FLOAT(), nullable=True)
        )
        batch_op.add_column(sa.Column("slope_7", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("slope_14", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("ma_50_200_buy_date", sa.DATE(), nullable=True))
        batch_op.add_column(sa.Column("rsi_last_value", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("last_week_growth", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("year_to_date_growth", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("last_dmp", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("star_wow", sa.FLOAT(), nullable=True))
        batch_op.add_column(
            sa.Column("last_week_max_growth", sa.FLOAT(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("last_month_max_growth", sa.FLOAT(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("year_to_date_max_growth", sa.FLOAT(), nullable=True)
        )
        batch_op.add_column(sa.Column("last_year_growth", sa.FLOAT(), nullable=True))
        batch_op.drop_index("ix_analysis_website")
        batch_op.drop_index("ix_analysis_trailing_price_to_earnings")
        batch_op.drop_index("ix_analysis_trailing_earnings_per_share")
        batch_op.drop_index("ix_analysis_summary")
        batch_op.drop_index("ix_analysis_stoch_oversold")
        batch_op.drop_index("ix_analysis_stoch_overbought")
        batch_op.drop_index("ix_analysis_sector")
        batch_op.drop_index("ix_analysis_rsi_oversold")
        batch_op.drop_index("ix_analysis_rsi_overbought")
        batch_op.drop_index("ix_analysis_rsi_bullish_crossover")
        batch_op.drop_index("ix_analysis_rsi_bearish_crossover")
        batch_op.drop_index("ix_analysis_revenue_growth")
        batch_op.drop_index("ix_analysis_return_on_equity")
        batch_op.drop_index("ix_analysis_rate_of_change_7_4")
        batch_op.drop_index("ix_analysis_rate_of_change_7_12")
        batch_op.drop_index("ix_analysis_rate_of_change_30")
        batch_op.drop_index("ix_analysis_rate_of_change_1")
        batch_op.drop_index("ix_analysis_quarterly_positive_return_on_equity")
        batch_op.drop_index("ix_analysis_quarterly_positive_return_on_assets")
        batch_op.drop_index("ix_analysis_quarterly_positive_operating_income")
        batch_op.drop_index("ix_analysis_quarterly_positive_net_income")
        batch_op.drop_index("ix_analysis_quarterly_positive_free_cash_flow")
        batch_op.drop_index("ix_analysis_quarterly_positive_diluted_eps")
        batch_op.drop_index("ix_analysis_quarterly_positive_debt_to_equity")
        batch_op.drop_index("ix_analysis_quarterly_positive_basic_eps")
        batch_op.drop_index(
            "ix_analysis_quarterly_operating_cash_flow_is_higher_than_net_income"
        )
        batch_op.drop_index("ix_analysis_quarterly_min_dividend_payout_ratio")
        batch_op.drop_index("ix_analysis_quarterly_min_capex_ratio")
        batch_op.drop_index("ix_analysis_quarterly_mean_dividend_payout_ratio")
        batch_op.drop_index("ix_analysis_quarterly_mean_capex_ratio")
        batch_op.drop_index("ix_analysis_quarterly_max_dividend_payout_ratio")
        batch_op.drop_index("ix_analysis_quarterly_max_capex_ratio")
        batch_op.drop_index("ix_analysis_quarterly_growing_operating_income")
        batch_op.drop_index("ix_analysis_quarterly_growing_operating_cash_flow")
        batch_op.drop_index("ix_analysis_quarterly_growing_net_income")
        batch_op.drop_index("ix_analysis_quarterly_growing_diluted_eps")
        batch_op.drop_index("ix_analysis_quarterly_growing_basic_eps")
        batch_op.drop_index("ix_analysis_quarterly_earning_per_share")
        batch_op.drop_index("ix_analysis_price_to_book")
        batch_op.drop_index("ix_analysis_price_per_earning_ratio")
        batch_op.drop_index("ix_analysis_positive_return_on_equity")
        batch_op.drop_index("ix_analysis_positive_return_on_assets")
        batch_op.drop_index("ix_analysis_positive_operating_income")
        batch_op.drop_index("ix_analysis_positive_net_income")
        batch_op.drop_index("ix_analysis_positive_free_cash_flow")
        batch_op.drop_index("ix_analysis_positive_diluted_eps")
        batch_op.drop_index("ix_analysis_positive_debt_to_equity")
        batch_op.drop_index("ix_analysis_positive_basic_eps")
        batch_op.drop_index("ix_analysis_operating_margins")
        batch_op.drop_index("ix_analysis_operating_cash_flow_is_higher_than_net_income")
        batch_op.drop_index("ix_analysis_name")
        batch_op.drop_index("ix_analysis_modifier")
        batch_op.drop_index("ix_analysis_min_dividend_payout_ratio")
        batch_op.drop_index("ix_analysis_min_capex_ratio")
        batch_op.drop_index("ix_analysis_mfi_oversold")
        batch_op.drop_index("ix_analysis_mfi_overbought")
        batch_op.drop_index("ix_analysis_mean_dividend_payout_ratio")
        batch_op.drop_index("ix_analysis_mean_capex_ratio")
        batch_op.drop_index("ix_analysis_max_dividend_payout_ratio")
        batch_op.drop_index("ix_analysis_max_capex_ratio")
        batch_op.drop_index("ix_analysis_market_capitalization")
        batch_op.drop_index("ix_analysis_market")
        batch_op.drop_index("ix_analysis_macd_12_26_9_zero_line_cross_up")
        batch_op.drop_index("ix_analysis_macd_12_26_9_zero_line_cross_down")
        batch_op.drop_index("ix_analysis_macd_12_26_9_bullish_crossover")
        batch_op.drop_index("ix_analysis_macd_12_26_9_bearish_crossover")
        batch_op.drop_index("ix_analysis_last_price")
        batch_op.drop_index("ix_analysis_isin")
        batch_op.drop_index("ix_analysis_industry_group")
        batch_op.drop_index("ix_analysis_industry")
        batch_op.drop_index("ix_analysis_growing_operating_income")
        batch_op.drop_index("ix_analysis_growing_operating_cash_flow")
        batch_op.drop_index("ix_analysis_growing_net_income")
        batch_op.drop_index("ix_analysis_growing_diluted_eps")
        batch_op.drop_index("ix_analysis_growing_basic_eps")
        batch_op.drop_index("ix_analysis_gross_margins")
        batch_op.drop_index("ix_analysis_golden_cross")
        batch_op.drop_index("ix_analysis_forward_earnings_per_share")
        batch_op.drop_index("ix_analysis_exchange")
        batch_op.drop_index("ix_analysis_earning_per_share")
        batch_op.drop_index("ix_analysis_dividend_yield")
        batch_op.drop_index("ix_analysis_dividend_rate")
        batch_op.drop_index("ix_analysis_death_cross")
        batch_op.drop_index("ix_analysis_date")
        batch_op.drop_index("ix_analysis_currency")
        batch_op.drop_index("ix_analysis_created_at")
        batch_op.drop_index("ix_analysis_country")
        batch_op.drop_index("ix_analysis_cdltasukigap")
        batch_op.drop_index("ix_analysis_cdlpiercing")
        batch_op.drop_index("ix_analysis_cdlmorningstar")
        batch_op.drop_index("ix_analysis_cdlengulfing")
        batch_op.drop_index("ix_analysis_cdlabandonedbaby")
        batch_op.drop_index("ix_analysis_cdl3whitesoldiers")
        batch_op.drop_index("ix_analysis_cdl3linestrike")
        batch_op.drop_index("ix_analysis_book_value")
        batch_op.drop_index("ix_analysis_base_symbol")
        batch_op.drop_index("ix_analysis_adx_14_short")
        batch_op.drop_index("ix_analysis_adx_14_long")
        batch_op.drop_column("adx_14_short")
        batch_op.drop_column("adx_14_long")
        batch_op.drop_column("macd_12_26_9_zero_line_cross_down")
        batch_op.drop_column("macd_12_26_9_zero_line_cross_up")
        batch_op.drop_column("macd_12_26_9_bearish_crossover")
        batch_op.drop_column("macd_12_26_9_bullish_crossover")
        batch_op.drop_column("rsi_overbought")
        batch_op.drop_column("rsi_oversold")
        batch_op.drop_column("rsi_bearish_crossover")
        batch_op.drop_column("rsi_bullish_crossover")
        batch_op.drop_column("stoch_overbought")
        batch_op.drop_column("stoch_oversold")
        batch_op.drop_column("mfi_overbought")
        batch_op.drop_column("mfi_oversold")
        batch_op.drop_column("death_cross")
        batch_op.drop_column("golden_cross")
        batch_op.drop_column("rate_of_change_30")
        batch_op.drop_column("rate_of_change_7_12")
        batch_op.drop_column("rate_of_change_7_4")
        batch_op.drop_column("rate_of_change_1")
        batch_op.drop_column("cdlengulfing")
        batch_op.drop_column("cdlpiercing")
        batch_op.drop_column("cdltasukigap")
        batch_op.drop_column("cdlabandonedbaby")
        batch_op.drop_column("cdl3whitesoldiers")
        batch_op.drop_column("cdl3linestrike")
        batch_op.drop_column("cdlmorningstar")



import json
import logging
import re
import time
from datetime import date, datetime
from itertools import batched, chain
from pathlib import Path
from typing import (
    Annotated,
    Any,
    List,
    Optional,
    Sequence,
    Type,
    get_args,
    TYPE_CHECKING,
    ClassVar,
    Dict,
)

import pandas as pd
from bearish.models.assets.equity import BaseEquity  # type: ignore
from bearish.models.base import (  # type: ignore
    DataSourceBase,
    Ticker,
    PriceTracker,
    TrackerQuery,
    FinancialsTracker,
)
from bearish.models.financials.balance_sheet import (  # type: ignore
    BalanceSheet,
    QuarterlyBalanceSheet,
)
from bearish.models.financials.base import Financials, FinancialsWithDate  # type: ignore
from bearish.models.financials.cash_flow import (  # type: ignore
    CashFlow,
    QuarterlyCashFlow,
)
from bearish.models.financials.metrics import (  # type: ignore
    FinancialMetrics,
    QuarterlyFinancialMetrics,
)
from bearish.models.price.prices import Prices  # type: ignore
from bearish.models.query.query import AssetQuery, Symbols  # type: ignore
from bearish.types import TickerOnlySources  # type: ignore
from pydantic import BaseModel, BeforeValidator, Field, create_model

from bullish.analysis.indicators import Indicators, IndicatorModels, SignalSeries
from joblib import Parallel, delayed  # type: ignore

from bullish.analysis.industry_views import compute_industry_view

if TYPE_CHECKING:
    from bullish.database.crud import BullishDb

QUARTERLY = "quarterly"
logger = logging.getLogger(__name__)


def to_float(value: Any) -> Optional[float]:
    if value == "None":
        return None
    if value is None:
        return None
    if isinstance(value, str):
        try:
            return float(value)
        except ValueError:
            return None
    return float(value)


def _load_data(
    data: Sequence[DataSourceBase], symbol: str, class_: Type[DataSourceBase]
) -> pd.DataFrame:
    try:
        records = pd.DataFrame.from_records(
            [f.model_dump() for f in data if f.symbol == symbol]
        )
        return records.set_index("date").sort_index()
    except Exception as e:
        logger.warning(f"Failed to load data from {symbol}: {e}")
        columns = list(class_.model_fields)
        return pd.DataFrame(columns=columns).sort_index()


def _compute_growth(series: pd.Series) -> bool:
    if series.empty:
        return False
    return all(series.pct_change(fill_method=None).dropna() > 0)


def _all_positive(series: pd.Series, threshold: int = 0) -> bool:
    if series.empty:
        return False
    return all(series.dropna() > threshold)


def _get_last(data: pd.Series) -> Optional[float]:
    return data.iloc[-1] if not data.empty else None


def _abs(data: pd.Series) -> pd.Series:
    try:
        return abs(data)
    except Exception as e:
        logger.warning(f"Failed to compute absolute value: {e}")
        return data


class TechnicalAnalysisBase(BaseModel):
    _description: ClassVar[str] = "General technical indicators"
    last_price: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
        ),
    ]
    max_year_loss: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
        ),
    ]


TechnicalAnalysisModels = [*IndicatorModels, TechnicalAnalysisBase]


class TechnicalAnalysis(*TechnicalAnalysisModels):  # type: ignore

    @classmethod
    def from_data(cls, prices: pd.DataFrame, ticker: Ticker) -> "TechnicalAnalysis":
        if "close" not in prices.columns:
            logger.warning(
                f"Ticker {ticker.symbol} does not have valid 'close' values.",
                exc_info=True,
            )
            return cls()
        try:
            res = Indicators().compute(prices)
            last_price = prices.close.iloc[-1]
            max_price = prices.close.iloc[-253 * 2 :].max()
            max_year_loss = (max_price - last_price) / max_price
            return cls(last_price=last_price, max_year_loss=max_year_loss, **res)
        except Exception as e:
            logger.error(
                f"Failing to calculate technical analysis for {ticker.symbol}: {e}",
                exc_info=True,
            )
            return cls()


class BaseFundamentalAnalysis(BaseModel):
    positive_debt_to_equity: Optional[bool] = Field(
        None,
        description="True if the company's debt-to-equity ratio is favorable (typically low or improving).",
    )
    positive_return_on_assets: Optional[bool] = Field(
        None,
        description="True if the company reports a positive return on assets (ROA), "
        "indicating efficient use of its assets.",
    )
    positive_return_on_equity: Optional[bool] = Field(
        None,
        description="True if the return on equity (ROE) is positive, "
        "showing profitability relative to shareholder equity.",
    )
    positive_diluted_eps: Optional[bool] = Field(
        None,
        description="True if the diluted earnings per share (EPS), "
        "which includes the effect of convertible securities, is positive.",
    )
    positive_basic_eps: Optional[bool] = Field(
        None,
        description="True if the basic earnings per share (EPS) is positive, reflecting profitable operations.",
    )
    growing_basic_eps: Optional[bool] = Field(
        None,
        description="True if the basic EPS has shown consistent growth over a defined time period.",
    )
    growing_diluted_eps: Optional[bool] = Field(
        None,
        description="True if the diluted EPS has consistently increased over time.",
    )
    positive_net_income: Optional[bool] = Field(
        None,
        description="True if the net income is positive, indicating overall profitability.",
    )
    positive_operating_income: Optional[bool] = Field(
        None,
        description="True if the company has positive operating income from its core business operations.",
    )
    growing_net_income: Optional[bool] = Field(
        None, description="True if net income has shown consistent growth over time."
    )
    growing_operating_income: Optional[bool] = Field(
        None,
        description="True if the operating income has consistently increased over a period.",
    )
    positive_free_cash_flow: Optional[bool] = Field(
        None,
        description="True if the company has positive free cash flow, indicating financial flexibility and health.",
    )
    growing_operating_cash_flow: Optional[bool] = Field(
        None,
        description="True if the company's operating cash flow is growing steadily.",
    )
    operating_cash_flow_is_higher_than_net_income: Optional[bool] = Field(
        None,
        description="True if the operating cash flow exceeds net income, often a sign of high-quality earnings.",
    )

    # Capital Expenditure Ratios
    mean_capex_ratio: Optional[float] = Field(
        None,
        description="Average capital expenditure (CapEx) ratio, usually "
        "calculated as CapEx divided by revenue or operating cash flow.",
    )
    max_capex_ratio: Optional[float] = Field(
        None, description="Maximum observed CapEx ratio over the evaluation period."
    )
    min_capex_ratio: Optional[float] = Field(
        None, description="Minimum observed CapEx ratio over the evaluation period."
    )

    # Dividend Payout Ratios
    mean_dividend_payout_ratio: Optional[float] = Field(
        None,
        description="Average dividend payout ratio, representing the proportion of earnings paid out as dividends.",
    )
    max_dividend_payout_ratio: Optional[float] = Field(
        None, description="Maximum dividend payout ratio observed over the period."
    )
    min_dividend_payout_ratio: Optional[float] = Field(
        None, description="Minimum dividend payout ratio observed over the period."
    )

    # EPS Value
    earning_per_share: Optional[float] = Field(
        None,
        description="The latest or most relevant value of earnings per share (EPS), indicating net income per share.",
    )

    def is_empty(self) -> bool:
        return all(getattr(self, field) is None for field in self.model_fields)

    @classmethod
    def from_financials(
        cls, financials: "Financials", ticker: Ticker
    ) -> "BaseFundamentalAnalysis":
        return cls._from_financials(
            balance_sheets=financials.balance_sheets,
            financial_metrics=financials.financial_metrics,
            cash_flows=financials.cash_flows,
            ticker=ticker,
        )

    @classmethod
    def _from_financials(
        cls,
        balance_sheets: List[BalanceSheet] | List[QuarterlyBalanceSheet],
        financial_metrics: List[FinancialMetrics] | List[QuarterlyFinancialMetrics],
        cash_flows: List[CashFlow] | List[QuarterlyCashFlow],
        ticker: Ticker,
    ) -> "BaseFundamentalAnalysis":
        try:
            symbol = ticker.symbol

            balance_sheet = _load_data(balance_sheets, symbol, BalanceSheet)
            financial = _load_data(financial_metrics, symbol, FinancialMetrics)
            cash_flow = _load_data(cash_flows, symbol, CashFlow)

            # Debt-to-equity
            debt_to_equity = (
                balance_sheet.total_liabilities / balance_sheet.total_shareholder_equity
            ).dropna()
            positive_debt_to_equity = _all_positive(debt_to_equity, threshold=1)

            # Add relevant balance sheet data to financials
            financial["total_shareholder_equity"] = balance_sheet[
                "total_shareholder_equity"
            ]
            financial["common_stock_shares_outstanding"] = balance_sheet[
                "common_stock_shares_outstanding"
            ]

            # EPS and income checks
            earning_per_share = _get_last(
                (
                    financial.net_income / financial.common_stock_shares_outstanding
                ).dropna()
            )
            positive_net_income = _all_positive(financial.net_income)
            positive_operating_income = _all_positive(financial.operating_income)
            growing_net_income = _compute_growth(financial.net_income)
            growing_operating_income = _compute_growth(financial.operating_income)
            positive_diluted_eps = _all_positive(financial.diluted_eps)
            positive_basic_eps = _all_positive(financial.basic_eps)
            growing_basic_eps = _compute_growth(financial.basic_eps)
            growing_diluted_eps = _compute_growth(financial.diluted_eps)

            # Profitability ratios
            return_on_equity = (
                financial.net_income * 100 / financial.total_shareholder_equity
            ).dropna()
            return_on_assets = (
                financial.net_income * 100 / balance_sheet.total_assets
            ).dropna()
            positive_return_on_assets = _all_positive(return_on_assets)
            positive_return_on_equity = _all_positive(return_on_equity)
            # Cash flow analysis
            cash_flow["net_income"] = financial["net_income"]
            free_cash_flow = (
                cash_flow["operating_cash_flow"] - cash_flow["capital_expenditure"]
            )
            positive_free_cash_flow = _all_positive(free_cash_flow)
            growing_operating_cash_flow = _compute_growth(
                cash_flow["operating_cash_flow"]
            )
            operating_income_net_income = cash_flow[
                ["operating_cash_flow", "net_income"]
            ].dropna()
            operating_cash_flow_is_higher_than_net_income = all(
                operating_income_net_income["operating_cash_flow"]
                >= operating_income_net_income["net_income"]
            )
            cash_flow["capex_ratio"] = (
                cash_flow["capital_expenditure"] / cash_flow["operating_cash_flow"]
            ).dropna()
            mean_capex_ratio = cash_flow["capex_ratio"].mean()
            max_capex_ratio = cash_flow["capex_ratio"].max()
            min_capex_ratio = cash_flow["capex_ratio"].min()
            dividend_payout_ratio = (
                _abs(cash_flow["cash_dividends_paid"]) / free_cash_flow
            ).dropna()
            mean_dividend_payout_ratio = dividend_payout_ratio.mean()
            max_dividend_payout_ratio = dividend_payout_ratio.max()
            min_dividend_payout_ratio = dividend_payout_ratio.min()

            return cls(
                earning_per_share=earning_per_share,
                positive_debt_to_equity=positive_debt_to_equity,
                positive_return_on_assets=positive_return_on_assets,
                positive_return_on_equity=positive_return_on_equity,
                growing_net_income=growing_net_income,
                growing_operating_income=growing_operating_income,
                positive_diluted_eps=positive_diluted_eps,
                positive_basic_eps=positive_basic_eps,
                growing_basic_eps=growing_basic_eps,
                growing_diluted_eps=growing_diluted_eps,
                positive_net_income=positive_net_income,
                positive_operating_income=positive_operating_income,
                positive_free_cash_flow=positive_free_cash_flow,
                growing_operating_cash_flow=growing_operating_cash_flow,
                operating_cash_flow_is_higher_than_net_income=operating_cash_flow_is_higher_than_net_income,
                mean_capex_ratio=mean_capex_ratio,
                max_capex_ratio=max_capex_ratio,
                min_capex_ratio=min_capex_ratio,
                mean_dividend_payout_ratio=mean_dividend_payout_ratio,
                max_dividend_payout_ratio=max_dividend_payout_ratio,
                min_dividend_payout_ratio=min_dividend_payout_ratio,
            )
        except Exception as e:
            logger.error(
                f"Failed to compute fundamental analysis for {ticker}: {e}",
                exc_info=True,
            )
            return cls()


class YearlyFundamentalAnalysis(BaseFundamentalAnalysis): ...


fields_with_prefix = {
    f"{QUARTERLY}_{name}": (
        field_info.annotation,
        Field(default=None, description=field_info.description),
    )
    for name, field_info in BaseFundamentalAnalysis.model_fields.items()
}

# Create the new model
BaseQuarterlyFundamentalAnalysis = create_model(  # type: ignore
    "BaseQuarterlyFundamentalAnalysis", **fields_with_prefix
)


class QuarterlyFundamentalAnalysis(BaseQuarterlyFundamentalAnalysis):  # type: ignore
    @classmethod
    def from_quarterly_financials(
        cls, financials: "Financials", ticker: Ticker
    ) -> "QuarterlyFundamentalAnalysis":
        base_financial_analisys = BaseFundamentalAnalysis._from_financials(
            balance_sheets=financials.quarterly_balance_sheets,
            financial_metrics=financials.quarterly_financial_metrics,
            cash_flows=financials.quarterly_cash_flows,
            ticker=ticker,
        )
        return cls.model_validate({f"{QUARTERLY}_{k}": v for k, v in base_financial_analisys.model_dump().items()})  # type: ignore


class FundamentalAnalysis(YearlyFundamentalAnalysis, QuarterlyFundamentalAnalysis):
    @classmethod
    def from_financials(
        cls, financials: Financials, ticker: Ticker
    ) -> "FundamentalAnalysis":
        yearly_analysis = YearlyFundamentalAnalysis.from_financials(
            financials=financials, ticker=ticker
        )
        quarterly_analysis = QuarterlyFundamentalAnalysis.from_quarterly_financials(
            financials=financials, ticker=ticker
        )
        return FundamentalAnalysis.model_validate(
            yearly_analysis.model_dump() | quarterly_analysis.model_dump()
        )

    @classmethod
    def compute_series(
        cls, financials: FinancialsWithDate, ticker: Ticker
    ) -> List[SignalSeries]:
        fundamendal_analysis = FundamentalAnalysis.from_financials(financials, ticker)
        fundamental_analysis_ = fundamendal_analysis.model_dump(
            exclude_none=True, exclude_unset=True, exclude_defaults=True
        )
        fundamental_analysis_ = {
            k: v for k, v in fundamental_analysis_.items() if v is True
        }
        return [
            SignalSeries(
                name=k.upper(), symbol=ticker.symbol, value=v, date=financials.date
            )
            for k, v in fundamental_analysis_.items()
        ]


class AnalysisEarningsDate(BaseModel):
    next_earnings_date: Optional[date] = None


class AnalysisView(BaseModel):
    sector: Annotated[
        Optional[str],
        Field(
            None,
            description="Broad sector to which the company belongs, "
            "such as 'Real Estate' or 'Technology'",
        ),
    ]
    industry: Annotated[
        Optional[str],
        Field(
            None,
            description="Detailed industry categorization for the company, "
            "like 'Real Estate Management & Development'",
        ),
    ]
    market_capitalization: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Market capitalization value",
        ),
    ]
    country: Annotated[
        Optional[str],
        Field(None, description="Country where the company's headquarters is located"),
    ]
    symbol: str = Field(
        description="Unique ticker symbol identifying the company on the stock exchange"
    )
    name: Annotated[
        Optional[str],
        Field(None, description="Full name of the company"),
    ]
    price_per_earning_ratio: Optional[float] = None
    last_price: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
        ),
    ]
    max_year_loss: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
        ),
    ]
    consensus: Optional[str] = None
    recommendation: Optional[str] = None
    yearly_growth: Optional[float] = None
    weekly_growth: Optional[float] = None
    monthly_growth: Optional[float] = None
    upside: Optional[float] = None
    oai_high_price_target: Optional[float] = None
    oai_low_price_target: Optional[float] = None
    rsi: Optional[float] = None
    oai_recommendation: Optional[str] = None
    oai_moat: Optional[bool] = None


def json_loads(value: Any) -> Any:
    if isinstance(value, str):
        try:
            return json.loads(value)
        except Exception as e:
            logger.debug(e)
            return None
    return value


def scrub(text: str) -> str:
    strip_markup = re.compile(r"[\\`*_{}\[\]()>#+\-.!|~:$;\"\'<>&]").sub
    return strip_markup("", text)


class SubjectAnalysis(BaseModel):
    high_price_target: Optional[float] = None
    low_price_target: Optional[float] = None
    consensus: Optional[str] = None
    reason: Optional[str] = None
    recommendation: Optional[str] = None
    explanation: Optional[str] = None
    news_date: Optional[datetime] = None
    news_summary: Annotated[
        Optional[List[Dict[str, Any]]], BeforeValidator(json_loads)
    ] = None
    summary: Annotated[Optional[Dict[str, Any]], BeforeValidator(json_loads)] = None
    upside: Optional[float] = None
    downside: Optional[float] = None

    oai_high_price_target: Optional[float] = None
    oai_low_price_target: Optional[float] = None
    oai_news_date: Optional[datetime] = None
    oai_recent_news: Optional[str] = None
    oai_recommendation: Optional[str] = None
    oai_explanation: Optional[str] = None
    oai_moat: Optional[bool] = None

    def compute_upside(self, last_price: float) -> None:
        if self.oai_high_price_target is not None:
            self.upside = (
                (float(self.oai_high_price_target) - float(last_price))
                * 100
                / float(last_price)
            )
        if self.oai_low_price_target is not None:
            self.downside = (
                (float(last_price) - float(self.oai_low_price_target))
                * 100
                / float(last_price)
            )

    def to_news(self) -> Optional[str]:
        if not self.news_summary:
            return None
        return "".join(
            [
                f"<p>{t.get('content').replace("\n","")}</p>"  # type: ignore
                for t in self.news_summary
                if t.get("content")
            ]
        )

    def to_date(self) -> Optional[date]:
        if self.news_date:
            return self.news_date.date()
        return None


class Analysis(SubjectAnalysis, AnalysisEarningsDate, AnalysisView, BaseEquity, TechnicalAnalysis, FundamentalAnalysis):  # type: ignore

    @classmethod
    def from_ticker(cls, bearish_db: "BullishDb", ticker: Ticker) -> "Analysis":
        asset = bearish_db.read_assets(
            AssetQuery(
                symbols=Symbols(equities=[ticker]),
                excluded_sources=get_args(TickerOnlySources),
            )
        )

        equity = asset.get_one_equity()
        financials = Financials.from_ticker(bearish_db, ticker)
        fundamental_analysis = FundamentalAnalysis.from_financials(financials, ticker)
        prices = Prices.from_ticker(bearish_db, ticker)
        technical_analysis = TechnicalAnalysis.from_data(prices.to_dataframe(), ticker)
        next_earnings_date = bearish_db.read_next_earnings_date(ticker.symbol)
        subject = bearish_db.read_subject(ticker.symbol)
        if subject:
            subject.compute_upside(technical_analysis.last_price)

        return cls.model_validate(
            equity.model_dump()
            | fundamental_analysis.model_dump()
            | technical_analysis.model_dump()
            | (subject.model_dump() if subject else {})
            | {
                "next_earnings_date": next_earnings_date,
                "price_per_earning_ratio": (
                    (
                        technical_analysis.last_price
                        / fundamental_analysis.earning_per_share
                    )
                    if technical_analysis.last_price is not None
                    and fundamental_analysis.earning_per_share != 0
                    and fundamental_analysis.earning_per_share is not None
                    else None
                ),
            }
        )


def compute_financials_series(
    financials_: Financials, ticker: Ticker
) -> List[SignalSeries]:
    financials_with_dates = FinancialsWithDate.from_financials(financials_)
    series = []
    for f in financials_with_dates:
        series.extend(FundamentalAnalysis.compute_series(f, ticker))
    return series


def compute_analysis(database_path: Path, ticker: Ticker) -> Analysis:
    from bullish.database.crud import BullishDb

    bullish_db = BullishDb(database_path=database_path)
    return Analysis.from_ticker(bullish_db, ticker)


def compute_signal_series(database_path: Path, ticker: Ticker) -> List[SignalSeries]:
    from bullish.database.crud import BullishDb

    bullish_db = BullishDb(database_path=database_path)
    indicators = Indicators()
    prices = Prices.from_ticker(bullish_db, ticker)
    signal_series = indicators.compute_series(prices.to_dataframe(), ticker.symbol)
    financials = Financials.from_ticker(bullish_db, ticker)
    financial_series = compute_financials_series(financials, ticker)
    return signal_series + financial_series


def run_signal_series_analysis(bullish_db: "BullishDb") -> None:
    price_trackers = set(bullish_db._read_tracker(TrackerQuery(), PriceTracker))
    finance_trackers = set(bullish_db._read_tracker(TrackerQuery(), FinancialsTracker))
    tickers = list(price_trackers.intersection(finance_trackers))
    parallel = Parallel(n_jobs=-1)

    for batch_ticker in batched(tickers, 1):
        many_signal_series = parallel(
            delayed(compute_signal_series)(bullish_db.database_path, ticker)
            for ticker in batch_ticker
        )
        series = list(chain.from_iterable(many_signal_series))
        try:
            bullish_db.write_signal_series(series)
        except Exception as e:
            logger.error(f"Failed to compute signal series for {batch_ticker}: {e}")


def run_analysis(bullish_db: "BullishDb") -> None:
    compute_industry_view(bullish_db)
    price_trackers = set(bullish_db._read_tracker(TrackerQuery(), PriceTracker))
    finance_trackers = set(bullish_db._read_tracker(TrackerQuery(), FinancialsTracker))
    tickers = list(price_trackers.intersection(finance_trackers))
    parallel = Parallel(n_jobs=-1)

    for batch_ticker in batched(tickers, 1000):
        start = time.perf_counter()
        many_analysis = parallel(
            delayed(compute_analysis)(bullish_db.database_path, ticker)
            for ticker in batch_ticker
        )
        bullish_db.write_many_analysis(many_analysis)
        elapsed_time = time.perf_counter() - start
        print(
            f"Computed analysis for {len(batch_ticker)} tickers in {elapsed_time:.2f} seconds."
        )



"""

Revision ID: 49c83f9eb5ac
Revises: ee5baabb35f8
Create Date: 2025-07-06 07:29:51.181393

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "49c83f9eb5ac"
down_revision: Union[str, None] = "ee5baabb35f8"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("median_rate_of_change_1", sa.Float(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("median_rate_of_change_7_4", sa.Float(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("median_rate_of_change_7_12", sa.Float(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("median_rate_of_change_30", sa.Float(), nullable=True)
        )
        batch_op.add_column(sa.Column("rate_of_change_7", sa.Float(), nullable=True))
        batch_op.drop_index(batch_op.f("ix_analysis_rate_of_change_1"))
        batch_op.drop_index(batch_op.f("ix_analysis_rate_of_change_7_12"))
        batch_op.drop_index(batch_op.f("ix_analysis_rate_of_change_7_4"))
        batch_op.create_index(
            "ix_analysis_median_rate_of_change_1",
            ["median_rate_of_change_1"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_median_rate_of_change_30",
            ["median_rate_of_change_30"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_median_rate_of_change_7_12",
            ["median_rate_of_change_7_12"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_median_rate_of_change_7_4",
            ["median_rate_of_change_7_4"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_rate_of_change_7", ["rate_of_change_7"], unique=False
        )
        batch_op.drop_column("rate_of_change_7_4")
        batch_op.drop_column("rate_of_change_7_12")
        batch_op.drop_column("rate_of_change_1")

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("rate_of_change_1", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("rate_of_change_7_12", sa.FLOAT(), nullable=True))
        batch_op.add_column(sa.Column("rate_of_change_7_4", sa.FLOAT(), nullable=True))
        batch_op.drop_index("ix_analysis_rate_of_change_7")
        batch_op.drop_index("ix_analysis_median_rate_of_change_7_4")
        batch_op.drop_index("ix_analysis_median_rate_of_change_7_12")
        batch_op.drop_index("ix_analysis_median_rate_of_change_30")
        batch_op.drop_index("ix_analysis_median_rate_of_change_1")
        batch_op.create_index(
            batch_op.f("ix_analysis_rate_of_change_7_4"),
            ["rate_of_change_7_4"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_analysis_rate_of_change_7_12"),
            ["rate_of_change_7_12"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_analysis_rate_of_change_1"),
            ["rate_of_change_1"],
            unique=False,
        )
        batch_op.drop_column("rate_of_change_7")
        batch_op.drop_column("median_rate_of_change_30")
        batch_op.drop_column("median_rate_of_change_7_12")
        batch_op.drop_column("median_rate_of_change_7_4")
        batch_op.drop_column("median_rate_of_change_1")

    # ### end Alembic commands ###



import logging
import shelve
import uuid
from pathlib import Path
from typing import Optional, List, Type, Dict, Any

import pandas as pd
import streamlit as st

import streamlit_pydantic as sp
from bearish.models.base import Ticker  # type: ignore
from bearish.models.price.prices import Prices  # type: ignore
from bearish.models.query.query import AssetQuery, Symbols  # type: ignore
from streamlit_file_browser import st_file_browser  # type: ignore

from bullish.analysis.industry_views import get_industry_comparison_data
from bullish.analysis.predefined_filters import PredefinedFilters
from bullish.database.crud import BullishDb
from bullish.figures.figures import plot
from bullish.analysis.filter import (
    FilterQuery,
    FilterUpdate,
    FilteredResults,
    FilterQueryStored,
    FundamentalAnalysisFilters,
    GROUP_MAPPING,
    GeneralFilter,
    TechnicalAnalysisFilters,
)
from bullish.jobs.tasks import update, news, analysis, initialize
from pydantic import BaseModel

from bullish.utils.checks import (
    compatible_bearish_database,
    compatible_bullish_database,
    empty_analysis_table,
)

CACHE_SHELVE = "user_cache"
DB_KEY = "db_path"

st.set_page_config(page_title="Bullish", page_icon="💰", layout="wide")
logger = logging.getLogger(__name__)


@st.cache_resource
def db_id() -> str:
    return f"{DB_KEY}_{uuid.uuid4()!s}"


@st.cache_resource
def bearish_db(database_path: Path) -> BullishDb:
    return BullishDb(database_path=database_path)


def store_db(db_path: Path) -> None:
    with shelve.open(CACHE_SHELVE) as storage:  # noqa:S301
        storage[db_id()] = str(db_path)


def load_db() -> Optional[str]:
    with shelve.open(CACHE_SHELVE) as db:  # noqa:S301
        db_path = db.get(db_id())
        return db_path


def assign_db_state() -> None:
    if "database_path" not in st.session_state:
        st.session_state.database_path = load_db()


@st.cache_data(hash_funcs={BullishDb: lambda obj: hash(obj.database_path)})
def load_analysis_data(bullish_db: BullishDb) -> pd.DataFrame:
    return bullish_db.read_analysis_data()


def on_table_select() -> None:

    row = st.session_state.selected_data["selection"]["rows"]

    db = bearish_db(st.session_state.database_path)
    if st.session_state.data.empty or (
        not st.session_state.data.iloc[row]["symbol"].to_numpy()
    ):
        return

    symbol = st.session_state.data.iloc[row]["symbol"].to_numpy()[0]
    country = st.session_state.data.iloc[row]["country"].to_numpy()[0]
    industry = st.session_state.data.iloc[row]["industry"].to_numpy()[0]
    query = AssetQuery(symbols=Symbols(equities=[Ticker(symbol=symbol)]))
    prices = db.read_series(query, months=24)
    data = Prices(prices=prices).to_dataframe()
    dates = db.read_dates(symbol)
    subject = db.read_subject(symbol)
    industry_data = get_industry_comparison_data(db, data, "Mean", industry, country)

    fig = plot(data, symbol, dates=dates, industry_data=industry_data)

    st.session_state.ticker_figure = fig
    st.session_state.ticker_news = subject


@st.dialog("🔑  Select database file to continue")
def dialog_pick_database() -> None:
    current_working_directory = Path.cwd()
    event = st_file_browser(
        path=current_working_directory, key="A", glob_patterns="**/*.db"
    )
    if event:
        db_path = Path(current_working_directory).joinpath(event["target"]["path"])
        if not (db_path.exists() and db_path.is_file()):
            st.stop()
        if not compatible_bearish_database(db_path):
            st.error(f"The database {db_path} is not compatible with this application.")
            st.stop()
        st.session_state.database_path = db_path
        store_db(db_path)
        compatible_bullish_db = compatible_bullish_database(db_path)
        if (not compatible_bullish_db) or (
            compatible_bullish_db and empty_analysis_table(db_path)
        ):
            st.warning(
                f"The database {db_path} has not the necessary data to run this application. "
                "A backround job will be started to update the data."
            )
            analysis(db_path, "Update analysis")
        st.rerun()
    if event is None:
        st.stop()


@st.cache_resource
def symbols() -> List[str]:
    bearish_db_ = bearish_db(st.session_state.database_path)
    return bearish_db_.read_symbols()


def groups_mapping() -> Dict[str, List[str]]:
    GROUP_MAPPING["symbol"] = symbols()
    return GROUP_MAPPING


def build_filter(model: Type[BaseModel], data: Dict[str, Any]) -> Dict[str, Any]:

    for field, info in model.model_fields.items():
        name = info.description or info.alias or field
        default = info.default
        if data.get(field) and data[field] != info.default:
            default = data[field]
        if info.annotation == Optional[List[str]]:  # type: ignore
            mapping = groups_mapping().get(field)
            if not mapping:
                continue
            data[field] = st.multiselect(
                name,
                mapping,
                default=default,
                key=hash((model.__name__, field)),
            )
        elif info.annotation == Optional[str]:  # type: ignore
            mapping = groups_mapping().get(field)
            if not mapping:
                continue
            options = ["", *mapping]
            data[field] = st.selectbox(
                name,
                options,
                index=0 if not default else options.index(default),
                key=hash((model.__name__, field)),
            )

        else:
            ge = next(
                (item.ge for item in info.metadata if hasattr(item, "ge")),
                info.default[0] if info.default and len(info.default) == 2 else None,
            )
            le = next(
                (item.le for item in info.metadata if hasattr(item, "le")),
                info.default[1] if info.default and len(info.default) == 2 else None,
            )
            if info.annotation == Optional[List[float]]:  # type: ignore
                ge = int(ge)  # type: ignore
                le = int(le)  # type: ignore
                default = [int(d) for d in default]
            try:
                data[field] = list(
                    st.slider(  # type: ignore
                        name, ge, le, tuple(default), key=hash((model.__name__, field))
                    )
                )
            except Exception as e:
                logger.error(
                    f"Error building filter for {model.__name__}.{field} "
                    f"with the parameters {(info.annotation, name, ge, le)}: {e}"
                )
                raise e
    return data


@st.dialog("⏳  Jobs", width="large")
def jobs() -> None:
    with st.expander("Update data"):
        update_query = sp.pydantic_form(key="update", model=FilterUpdate)
        if (
            update_query
            and st.session_state.data is not None
            and not st.session_state.data.empty
        ):
            symbols = st.session_state.data["symbol"].unique().tolist()
            update(
                database_path=st.session_state.database_path,
                job_type="Update data",
                symbols=symbols,
                update_query=update_query,
            )  # enqueue & get result-handle

            st.success("Data update job has been enqueued.")
            st.rerun()


@st.dialog("📥  Load", width="large")
def load() -> None:
    bearish_db_ = bearish_db(st.session_state.database_path)
    existing_filtered_results = bearish_db_.read_list_filtered_results()
    option = st.selectbox("Select portfolio", ["", *existing_filtered_results])
    if option:
        filtered_results_ = bearish_db_.read_filtered_results(option)
        if filtered_results_:
            st.session_state.data = bearish_db_.read_analysis_data(
                symbols=filtered_results_.symbols
            )
            st.rerun()


@st.dialog("🔍  Filter", width="large")
def filter() -> None:
    with st.container(), st.expander("Predefined filters"):
        predefined_filter_names = PredefinedFilters().get_predefined_filter_names()
        option = st.selectbox(
            "Select a predefined filter",
            ["", *predefined_filter_names],
        )
        if option:
            data_ = PredefinedFilters().get_predefined_filter(option)
            st.session_state.filter_query.update(data_)
    with st.container():

        with st.expander("Technical Analysis"):
            for filter in TechnicalAnalysisFilters:
                with st.expander(filter._description):  # type: ignore
                    build_filter(filter, st.session_state.filter_query)

        with st.expander("Fundamental Analysis"):
            for filter in FundamentalAnalysisFilters:
                with st.expander(filter._description):  # type: ignore
                    build_filter(filter, st.session_state.filter_query)
        with st.expander("General filter"):
            build_filter(GeneralFilter, st.session_state.filter_query)

    if st.button("🔍 Apply"):
        query = FilterQuery.model_validate(st.session_state.filter_query)
        if query.valid():
            st.session_state.data = bearish_db(
                st.session_state.database_path
            ).read_filter_query(query)
            st.session_state.ticker_figure = None
            st.session_state.ticker_news = None
            st.session_state.filter_query = {}
            st.session_state.query = query
            st.rerun()


@st.dialog("📈  Price history and analysis", width="large")
def dialog_plot_figure() -> None:
    st.markdown(
        """
    <style>
    div[data-testid="stDialog"] div[role="dialog"]:has(.big-dialog) {
        width: 90vw;
        height: 170vh;
    }
    </style>
    """,
        unsafe_allow_html=True,
    )
    st.html("<span class='big-dialog'></span>")
    if st.session_state.ticker_news:
        st.markdown(
            f"""
            <div class="news-hover" >
              📰 <span class="label">News</span>
              <div class="tooltip">
                <h2>Date: {st.session_state.ticker_news.to_date()}</h2>
                <h2>Price targets</h2>
                <p>High price target: {st.session_state.ticker_news.high_price_target}</p>
                <p>Low price target: {st.session_state.ticker_news.low_price_target}</p>
                <p>OpenAI High price target: {st.session_state.ticker_news.oai_high_price_target}</p>
                <p>OpenAI Low price target: {st.session_state.ticker_news.oai_low_price_target}</p>
                <h2>Recommendation: {st.session_state.ticker_news.recommendation}</h2>
                <h2>OpenAI Recommendation: {st.session_state.ticker_news.oai_recommendation}</h2>
                <h2>Consensus: {st.session_state.ticker_news.consensus}</h2>
                <h2>Explanation & reasons</h2>
                <p>{st.session_state.ticker_news.explanation}</p>
                <p>{st.session_state.ticker_news.reason}</p>
                <p>{st.session_state.ticker_news.oai_explanation}</p>
                <h2>Recent news</h2>
                <p>{st.session_state.ticker_news.oai_recent_news}</p>
                <h2>News summaries</h2>
                {st.session_state.ticker_news.to_news()}
              </div>
            </div>    
            <style>
              /* Hover target (fixed top-left) */
              .news-hover {{
                position: absolute;
                left: 1rem;
                display: inline-flex;
                align-items: center;
                gap: .4rem;
                font-size: 1.7rem;       /* big label */
                font-weight: 600;
                color: #333;
                cursor: pointer;
                user-select: none;
                z-index: 1100;
              }}    
              /* Tooltip bubble */
              .news-hover .tooltip {{
                position: absolute;
                top: 110%;               /* below the label */
                left: 0;
                width: 840px;
                max-height: 620px;
                overflow-y: auto;
                background: #222;
                color: #fff;
                padding: 1.2rem;
                border-radius: 10px;
                font-size: .95rem;
                line-height: 1.45;
                box-shadow: 0 8px 20px rgba(0,0,0,.4);
                opacity: 0;
                visibility: hidden;
                transition: opacity .25s ease;
              }}
              .news-hover .tooltip hr {{
                border: none;
                border-top: 1px solid #444;
                margin: 1rem 0;
              }}    
              /* Show tooltip on hover or keyboard focus */
              .news-hover:hover .tooltip,
              .news-hover:focus-within .tooltip {{
                opacity: 1;
                visibility: visible;
              }}
              /* Little arrow under the bubble */
              .news-hover .tooltip::after {{
                content: "";
                position: absolute;
                top: -10px;
                left: 20px;
                border: 10px solid transparent;
                border-top-color: #222;
              }}
            </style>
            """,
            unsafe_allow_html=True,
        )

    st.plotly_chart(st.session_state.ticker_figure, use_container_width=True)
    st.session_state.ticker_figure = None
    st.session_state.ticker_news = None


@st.dialog("⭐ Save filtered results")
def save_filtered_results(bearish_db_: BullishDb) -> None:
    user_input = st.text_input("Selection name").strip()
    headless = st.checkbox("Headless mode", value=True)
    apply = st.button("Apply")
    if apply:
        if not user_input:
            st.error("This field is required.")
        else:
            symbols = st.session_state.data["symbol"].unique().tolist()
            filtered_results = FilteredResults(
                name=user_input,
                filter_query=FilterQueryStored.model_validate(
                    st.session_state.query.model_dump(
                        exclude_unset=True, exclude_defaults=True
                    )
                ),
                symbols=symbols,
            )

            bearish_db_.write_filtered_results(filtered_results)
            news(
                database_path=st.session_state.database_path,
                job_type="Fetching news",
                symbols=symbols,
                headless=headless,
            )
            st.session_state.filter_query = None
            st.session_state.query = None
            st.rerun()


def main() -> None:
    hide_elements = """
            <style>
                div[data-testid="stSliderTickBarMin"],
                div[data-testid="stSliderTickBarMax"] {
                    display: none;
                }
            </style>
    """

    st.markdown(hide_elements, unsafe_allow_html=True)
    assign_db_state()

    if st.session_state.database_path is None:
        dialog_pick_database()
    if "initialized" not in st.session_state:
        initialize(
            database_path=st.session_state.database_path,
            job_type="Initialize",
        )
        st.session_state.initialized = True
    bearish_db_ = bearish_db(st.session_state.database_path)

    charts_tab, jobs_tab = st.tabs(["Charts", "Jobs"])
    if "data" not in st.session_state:
        st.session_state.data = load_analysis_data(bearish_db_)

    with charts_tab:
        with st.container():
            columns = st.columns(12)
            with columns[0]:
                if st.button(" 🔍 ", use_container_width=True):
                    st.session_state.filter_query = {}
                    filter()
            with columns[1]:
                if (
                    "query" in st.session_state
                    and st.session_state.query is not None
                    and st.session_state.query.valid()
                ):
                    favorite = st.button(" ⭐ ", use_container_width=True)
                    if favorite:
                        save_filtered_results(bearish_db_)
            with columns[-1]:
                if st.button(" 📥 ", use_container_width=True):
                    load()

        with st.container():
            st.dataframe(
                st.session_state.data,
                on_select=on_table_select,
                selection_mode="single-row",
                key="selected_data",
                use_container_width=True,
                height=600,
            )
            if (
                "ticker_figure" in st.session_state
                and st.session_state.ticker_figure is not None
            ):
                dialog_plot_figure()

    with jobs_tab:
        columns = st.columns(12)
        with columns[0]:
            if st.button(" ⏳ ", use_container_width=True):
                jobs()

        job_trackers = bearish_db_.read_job_trackers()
        st.dataframe(
            job_trackers,
            use_container_width=True,
            hide_index=True,
        )


if __name__ == "__main__":
    main()



"""

Revision ID: bf6b86dd5463
Revises: 17e51420e7ad
Create Date: 2025-07-11 18:32:21.450156

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "bf6b86dd5463"
down_revision: Union[str, None] = "17e51420e7ad"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("momentum", sa.Float(), nullable=True))
        batch_op.create_index("ix_analysis_momentum", ["momentum"], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_momentum")
        batch_op.drop_column("momentum")

    # ### end Alembic commands ###



"""

Revision ID: c828e29e1105
Revises: b36c310f49ec
Create Date: 2025-08-15 17:57:09.541454

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "c828e29e1105"
down_revision: Union[str, None] = "b36c310f49ec"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("oai_high_price_target", sa.Float(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("oai_low_price_target", sa.Float(), nullable=True)
        )
        batch_op.add_column(sa.Column("oai_news_date", sa.DateTime(), nullable=True))
        batch_op.add_column(
            sa.Column(
                "oai_recent_news", sqlmodel.sql.sqltypes.AutoString(), nullable=True
            )
        )
        batch_op.add_column(
            sa.Column(
                "oai_recommendation", sqlmodel.sql.sqltypes.AutoString(), nullable=True
            )
        )
        batch_op.add_column(
            sa.Column(
                "oai_explanation", sqlmodel.sql.sqltypes.AutoString(), nullable=True
            )
        )
        batch_op.create_index(
            "ix_analysis_oai_explanation", ["oai_explanation"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_oai_high_price_target", ["oai_high_price_target"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_oai_low_price_target", ["oai_low_price_target"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_oai_news_date", ["oai_news_date"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_oai_recent_news", ["oai_recent_news"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_oai_recommendation", ["oai_recommendation"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_oai_recommendation")
        batch_op.drop_index("ix_analysis_oai_recent_news")
        batch_op.drop_index("ix_analysis_oai_news_date")
        batch_op.drop_index("ix_analysis_oai_low_price_target")
        batch_op.drop_index("ix_analysis_oai_high_price_target")
        batch_op.drop_index("ix_analysis_oai_explanation")
        batch_op.drop_column("oai_explanation")
        batch_op.drop_column("oai_recommendation")
        batch_op.drop_column("oai_recent_news")
        batch_op.drop_column("oai_news_date")
        batch_op.drop_column("oai_low_price_target")
        batch_op.drop_column("oai_high_price_target")

    # ### end Alembic commands ###



import datetime
from typing import Optional, List

import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from bullish.analysis.functions import (
    add_indicators,
    support_resistance,
)
from datetime import date


def plot(
    data: pd.DataFrame,
    symbol: str,
    name: Optional[str] = None,
    dates: Optional[List[date]] = None,
    industry_data: Optional[pd.DataFrame] = None,
) -> go.Figure:
    data = add_indicators(data)
    supports = support_resistance(data)

    fig = make_subplots(
        rows=7,
        cols=1,
        shared_xaxes=True,
        vertical_spacing=0.05,
        specs=[
            [{"rowspan": 2}],  # Row 1: main chart
            [None],  # Row 2: skipped (part of row 1)
            [{}],  # Row 3: RSI
            [{}],  # Row 4: MACD
            [{}],  # Row 5: ADX
            [{}],  # Row 6: OBV
            [{}],  # Row 7: ATR
        ],
        subplot_titles=(
            f"Price + SMAs ({symbol} [{name}])",
            f"RSI ({symbol} [{name}])",
            f"MACD ({symbol} [{name}])",
            f"ADX ({symbol} [{name}])",
            f"ATR ({symbol} [{name}])",
            f"ADOSC ({symbol} [{name}])",
        ),
    )
    # Row 1: Candlestick + SMAs
    fig.add_trace(
        go.Candlestick(
            x=data.index,
            open=data.open,
            high=data.high,
            low=data.low,
            close=data.close,
            name="Candlestick",
        ),
        row=1,
        col=1,
    )
    fig.update_xaxes(rangeslider_thickness=0.04, row=1, col=1)
    fig.add_trace(
        go.Scatter(x=data.index, y=data.SMA_50, name="SMA 50", mode="lines"),
        row=1,
        col=1,
    )

    fig.add_trace(
        go.Scatter(x=data.index, y=data.SMA_200, name="SMA 200", mode="lines"),
        row=1,
        col=1,
    )

    # Row 2: RSI
    fig.add_trace(
        go.Scatter(x=data.index, y=data.RSI, name="RSI 14", mode="lines"),
        row=3,
        col=1,
    )

    # Row 3: MACD
    fig.add_trace(
        go.Scatter(x=data.index, y=data.MACD_12_26_9, name="MACD", mode="lines"),
        row=4,
        col=1,
    )

    fig.add_trace(
        go.Scatter(
            x=data.index, y=data.MACD_12_26_9_SIGNAL, name="MACD Signal", mode="lines"
        ),
        row=4,
        col=1,
    )

    fig.add_trace(
        go.Bar(
            x=data.index, y=data.MACD_12_26_9_HIST, name="MACD Histogram", opacity=0.5
        ),
        row=4,
        col=1,
    )

    # Row 4: ADX
    fig.add_trace(
        go.Scatter(x=data.index, y=data.ADX_14, name="ADX_14", mode="lines"),
        row=5,
        col=1,
    )

    fig.add_trace(
        go.Scatter(x=data.index, y=data.MINUS_DI, name="-DI", mode="lines"),
        row=5,
        col=1,
    )
    fig.add_trace(
        go.Scatter(x=data.index, y=data.PLUS_DI, name="+DI", mode="lines"),
        row=5,
        col=1,
    )
    fig.add_trace(
        go.Scatter(x=data.index, y=data.ATR, name="ATR", mode="lines"),
        row=6,
        col=1,
    )
    fig.add_trace(
        go.Scatter(x=data.index, y=data.ADOSC, name="ADOSC", mode="lines"),
        row=7,
        col=1,
    )

    if dates is not None and dates:
        for date in dates:
            if (
                data.first_valid_index().date() > date  # type: ignore
                or data.last_valid_index().date() + datetime.timedelta(days=31 * 3)  # type: ignore
                < date
            ):
                continue
            fig.add_vline(
                x=date,
                line_dash="dashdot",
                line_color="MediumPurple",
                line_width=1,
                row=1,
                col=1,
            )

    # Layout tweaks
    fig.update_layout(
        height=1500,
        showlegend=True,
        title="Technical Indicator Dashboard",
        margin={"t": 60, "b": 40},
    )

    # Optional: Add horizontal lines for RSI (e.g., 70/30 levels)
    fig.add_hline(y=70, line_dash="dash", line_color="red", row=3, col=1)
    fig.add_hline(y=30, line_dash="dash", line_color="green", row=3, col=1)
    fig.add_hline(y=25, line_dash="dash", line_color="red", row=5, col=1)
    fig.add_hline(
        y=supports.support.value,
        line_dash="dash",
        line_color="rgba(26, 188, 156, 1)",  # teal, fully opaque
        annotation_text=f"Support ({supports.support.value:.2f})",
        line_width=0.75,
        row=1,
        col=1,
    )
    fig.add_hline(
        y=supports.resistance.value,
        line_dash="dash",
        line_color="rgba(230, 126, 34, 1)",  # orange, fully opaque
        annotation_text=f"Resistance ({supports.resistance.value:.2f})",
        line_width=0.75,
        row=1,
        col=1,
    )

    return fig



import datetime
from datetime import date
from typing import get_args, Any, Optional, List, Tuple, Type, Dict

from bearish.types import SeriesLength  # type: ignore
from pydantic import BaseModel, Field, ConfigDict
from pydantic import create_model
from pydantic.fields import FieldInfo

from bullish.analysis.analysis import (
    YearlyFundamentalAnalysis,
    QuarterlyFundamentalAnalysis,
    TechnicalAnalysisModels,
    TechnicalAnalysis,
    AnalysisView,
)
from bullish.analysis.constants import Industry, IndustryGroup, Sector, Country

SIZE_RANGE = 2


def _get_type(name: str, info: FieldInfo) -> Tuple[Any, Any]:
    alias = info.alias or " ".join(name.capitalize().split("_")).strip()
    if info.annotation == Optional[float]:  # type: ignore
        ge = next((item.ge for item in info.metadata if hasattr(item, "ge")), 0)
        le = next((item.le for item in info.metadata if hasattr(item, "le")), 100)
        default = [ge, le]
        return (
            Optional[List[float]],
            Field(default=default, alias=alias, description=info.description),
        )
    elif info.annotation == Optional[date]:  # type: ignore
        le = date.today()
        ge = le - datetime.timedelta(days=30 * 2)  # 30 days * 12 months
        return (
            List[date],
            Field(default=[ge, le], alias=alias, description=info.description),
        )
    else:
        raise NotImplementedError


FUNDAMENTAL_ANALYSIS_GROUP = ["income", "cash_flow", "eps"]


def _get_fundamental_analysis_boolean_fields() -> List[str]:
    return [
        name
        for name, info in {
            **YearlyFundamentalAnalysis.model_fields,
            **QuarterlyFundamentalAnalysis.model_fields,
        }.items()
        if info.annotation == Optional[bool]
    ]


def _get_technical_analysis_float_fields() -> List[str]:
    return [
        name
        for name, info in (
            TechnicalAnalysis.model_fields | AnalysisView.model_fields
        ).items()
        if info.annotation == Optional[float]
    ]


def get_boolean_field_group(group: str) -> List[str]:
    groups = FUNDAMENTAL_ANALYSIS_GROUP.copy()
    groups.remove(group)
    return [
        name
        for name in _get_fundamental_analysis_boolean_fields()
        if group in name and not any(g in name for g in groups)
    ]


INCOME_GROUP = get_boolean_field_group("income")
CASH_FLOW_GROUP = get_boolean_field_group("cash_flow")
EPS_GROUP = get_boolean_field_group("eps")
PROPERTIES_GROUP = list(
    set(_get_fundamental_analysis_boolean_fields()).difference(
        {*INCOME_GROUP, *CASH_FLOW_GROUP, *EPS_GROUP}
    )
)
BOOLEAN_GROUP_MAPPING: Dict[str, List[str]] = {
    "income": INCOME_GROUP,
    "cash_flow": CASH_FLOW_GROUP,
    "eps": EPS_GROUP,
    "properties": PROPERTIES_GROUP,
}
GROUP_MAPPING: Dict[str, List[str]] = {
    **BOOLEAN_GROUP_MAPPING,
    "properties": PROPERTIES_GROUP,
    "country": list(get_args(Country)),
    "industry": list(get_args(Industry)),
    "industry_group": list(get_args(IndustryGroup)),
    "sector": list(get_args(Sector)),
    "symbol": [],
    "order_by_asc": _get_technical_analysis_float_fields(),
    "order_by_desc": _get_technical_analysis_float_fields(),
}


def _create_fundamental_analysis_models() -> List[Type[BaseModel]]:
    models = []
    boolean_fields = {
        "income": (Optional[List[str]], Field(default=None, description="Income")),
        "cash_flow": (
            Optional[List[str]],
            Field(default=None, description="Cash flow"),
        ),
        "eps": (
            Optional[List[str]],
            Field(default=None, description="Earnings per share"),
        ),
        "properties": (
            Optional[List[str]],
            Field(default=None, description="General properties"),
        ),
    }
    yearly_fields = {
        name: _get_type(name, info)
        for name, info in YearlyFundamentalAnalysis.model_fields.items()
        if info.annotation != Optional[bool]  # type: ignore
    }
    quarterly_fields = {
        name: _get_type(name, info)
        for name, info in QuarterlyFundamentalAnalysis.model_fields.items()
        if info.annotation != Optional[bool]
    }
    for property in [
        (boolean_fields, "Selection filter", "SelectionFilter"),
        (yearly_fields, "Yearly properties", "YearlyFilter"),
        (quarterly_fields, "Quarterly properties", "QuarterlyFilter"),
    ]:
        model_ = create_model(  # type: ignore
            property[-1],
            __config__=ConfigDict(populate_by_name=True),
            **property[0],
        )
        model_._description = property[1]
        models.append(model_)

    return models


def create_technical_analysis_models() -> List[Type[BaseModel]]:
    models = []
    for model in TechnicalAnalysisModels:
        model_ = create_model(  # type: ignore
            f"{model.__name__}Filter",  # type: ignore
            __config__=ConfigDict(populate_by_name=True),
            **{
                name: _get_type(name, info) for name, info in model.model_fields.items()  # type: ignore
            },
        )

        model_._description = model._description  # type: ignore
        models.append(model_)
    return models


TechnicalAnalysisFilters = create_technical_analysis_models()
FundamentalAnalysisFilters = _create_fundamental_analysis_models()


class GeneralFilter(BaseModel):
    country: Optional[List[str]] = None
    order_by_asc: Optional[str] = None
    order_by_desc: Optional[str] = None
    industry: Optional[List[str]] = None
    industry_group: Optional[List[str]] = None
    sector: Optional[List[str]] = None
    symbol: Optional[List[str]] = None
    limit: Optional[str] = None
    next_earnings_date: List[date] = Field(
        default=[date.today(), date.today() + datetime.timedelta(days=30 * 12)],
    )
    market_capitalization: Optional[List[float]] = Field(default=[5e8, 1e12])
    price_per_earning_ratio: Optional[List[float]] = Field(default=[0.0, 1000.0])


class FilterQuery(GeneralFilter, *TechnicalAnalysisFilters, *FundamentalAnalysisFilters):  # type: ignore

    def valid(self) -> bool:
        return any(
            bool(v)
            for _, v in self.model_dump(
                exclude_defaults=True, exclude_unset=True
            ).items()
        )

    def to_query(self) -> str:  # noqa: C901
        parameters = self.model_dump(
            exclude_defaults=True, exclude_unset=True, exclude={"name"}
        )
        query = []
        order_by_desc = ""
        order_by_asc = ""
        limit = None
        for parameter, value in parameters.items():
            if not value:
                continue

            if (
                isinstance(value, list)
                and all(isinstance(item, str) for item in value)
                and parameter not in GeneralFilter.model_fields
            ):
                query.append(" AND ".join([f"{v}=1" for v in value]))
            elif (
                isinstance(value, str) and bool(value) and parameter == "order_by_desc"
            ):
                order_by_desc = f"ORDER BY {value} DESC"
            elif isinstance(value, str) and bool(value) and parameter == "order_by_asc":
                order_by_asc = f"ORDER BY {value} ASC"
            elif isinstance(value, str) and bool(value) and parameter == "limit":
                limit = f" LIMIT {int(value)}"
            elif (
                isinstance(value, list)
                and len(value) == SIZE_RANGE
                and all(isinstance(item, (int, float)) for item in value)
            ):
                query.append(f"{parameter} BETWEEN {value[0]} AND {value[1]}")
            elif (
                (
                    isinstance(value, list)
                    and len(value) == SIZE_RANGE
                    and all(isinstance(item, date) for item in value)
                )
                and parameter == "next_earnings_date"
                or (
                    isinstance(value, list)
                    and len(value) == SIZE_RANGE
                    and all(isinstance(item, date) for item in value)
                )
            ):
                query.append(f"{parameter} BETWEEN '{value[0]}' AND '{value[1]}'")
            elif (
                isinstance(value, list)
                and all(isinstance(item, str) for item in value)
                and parameter in GeneralFilter.model_fields
            ):
                general_filters = [f"'{v}'" for v in value]
                query.append(f"{parameter} IN ({', '.join(general_filters)})")
            else:
                raise NotImplementedError(f"with value {value}")
        query_ = " AND ".join(query)
        query__ = f"{query_} {order_by_desc.strip()} {order_by_asc.strip()}".strip()
        if limit is not None:
            query__ += limit
        else:
            query__ += " LIMIT 1000"
        return query__


class FilterQueryStored(FilterQuery): ...


class FilterUpdate(BaseModel):
    window_size: SeriesLength = Field("1mo")
    data_age_in_days: int = 1
    update_financials: bool = False
    update_analysis_only: bool = False


class FilteredResults(BaseModel):
    name: str
    filter_query: FilterQueryStored
    symbols: list[str] = Field(
        default_factory=list, description="List of filtered tickers."
    )



"""

Revision ID: b36c310f49ec
Revises: 260fcff7212e
Create Date: 2025-08-14 22:39:38.207093

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "b36c310f49ec"
down_revision: Union[str, None] = "cc28171c21a4"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "openai",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("news_date", sa.Date(), nullable=False),
        sa.Column("high_price_target", sa.Float(), nullable=True),
        sa.Column("low_price_target", sa.Float(), nullable=True),
        sa.Column("recent_news", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("recommendation", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("explanation", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "news_date"),
    )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("openai")
    # ### end Alembic commands ###



"""

Revision ID: 79bc71ec6f9e
Revises: 6d252e23f543
Create Date: 2025-08-02 15:07:58.266933

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "79bc71ec6f9e"
down_revision: Union[str, None] = "6d252e23f543"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("max_year_loss", sa.Float(), nullable=True))
        batch_op.create_index(
            "ix_analysis_max_year_loss", ["max_year_loss"], unique=False
        )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_max_year_loss")
        batch_op.drop_column("max_year_loss")

    # ### end Alembic commands ###



import functools
import logging
from typing import Optional, Any, Callable, List

import pandas as pd
from bearish.main import Bearish  # type: ignore
from bearish.models.sec.sec import Secs  # type: ignore
from tickermood.main import get_news  # type: ignore
from tickermood.types import DatabaseConfig  # type: ignore

from .app import huey
from pathlib import Path
from huey.api import Task, crontab  # type: ignore

from .models import JobTrackerStatus, JobTracker, JobType
from ..analysis.analysis import run_analysis, run_signal_series_analysis
from ..analysis.backtest import run_many_tests, BackTestConfig
from ..analysis.industry_views import compute_industry_view
from ..analysis.openai import get_open_ai_news
from ..analysis.predefined_filters import predefined_filters, load_custom_filters
from ..database.crud import BullishDb
from bullish.analysis.filter import FilterUpdate

logger = logging.getLogger(__name__)


class DataBaseSingleTon:
    _instance = None

    def __new__(cls, *args: Any, **kwargs: Any) -> Any:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, path: Optional[Path] = None) -> None:
        if not hasattr(self, "path"):  # Only set once
            self.path = path

    def valid(self) -> bool:
        return hasattr(self, "path") and self.path is not None


def job_tracker(func: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(func)
    def wrapper(
        database_path: Path,
        job_type: JobType,
        *args: Any,
        task: Optional[Task] = None,
        **kwargs: Any,
    ) -> None:
        bullish_db = BullishDb(database_path=database_path)
        if task is None:
            raise ValueError("Task must be provided for job tracking.")
        if bullish_db.read_job_tracker(task.id) is None:
            bullish_db.write_job_tracker(JobTracker(job_id=str(task.id), type=job_type))
        bullish_db.update_job_tracker_status(
            JobTrackerStatus(job_id=task.id, status="Running")
        )
        try:
            func(database_path, job_type, *args, task=task, **kwargs)
            bullish_db.update_job_tracker_status(
                JobTrackerStatus(job_id=task.id, status="Completed")
            )
        except Exception as e:
            logger.exception(f"Fail to complete job {func.__name__}: {e}")
            bullish_db.update_job_tracker_status(
                JobTrackerStatus(job_id=task.id, status="Failed")
            )

    return wrapper


def _base_update(
    database_path: Path,
    job_type: JobType,
    symbols: Optional[List[str]],
    update_query: FilterUpdate,
    task: Optional[Task] = None,
) -> None:

    if not update_query.update_analysis_only:
        bearish = Bearish(path=database_path, auto_migration=False)
        bearish.update_prices(
            series_length=update_query.window_size,
            delay=update_query.data_age_in_days,
        )
        bearish.get_prices_index(series_length=update_query.window_size)
        Secs.upload(bearish._bearish_db)
        if update_query.update_financials:
            bearish.update_financials()
    bullish_db = BullishDb(database_path=database_path)
    run_analysis(bullish_db)
    compute_industry_view(bullish_db)


@huey.task(context=True)  # type: ignore
@job_tracker
def initialize(
    database_path: Path,
    job_type: JobType,
    task: Optional[Task] = None,
) -> None:
    database = DataBaseSingleTon(path=database_path)
    if not database.valid():
        raise ValueError("Database path is not valid.")


@huey.task(context=True)  # type: ignore
@job_tracker
def update(
    database_path: Path,
    job_type: JobType,
    symbols: Optional[List[str]],
    update_query: FilterUpdate,
    task: Optional[Task] = None,
) -> None:
    _base_update(database_path, job_type, symbols, update_query, task)


@huey.periodic_task(crontab(minute="0", hour="1"), context=True)  # type: ignore
def cron_update(
    task: Optional[Task] = None,
) -> None:
    database = DataBaseSingleTon()
    if database.valid():
        job_tracker(_base_update)(
            database.path, "Update data", [], FilterUpdate(), task=task
        )


@huey.periodic_task(crontab(day_of_week=0, hour=9, minute=0), context=True)  # type: ignore
def cron_financial_update(
    task: Optional[Task] = None,
) -> None:
    database = DataBaseSingleTon()
    if database.valid():
        job_tracker(_base_update)(
            database.path,
            "Update data",
            [],
            FilterUpdate(update_financials=True),
            task=task,
        )


@huey.task(context=True)  # type: ignore
@job_tracker
def analysis(
    database_path: Path,
    job_type: JobType,
    task: Optional[Task] = None,
) -> None:
    bullish_db = BullishDb(database_path=database_path)
    run_analysis(bullish_db)
    compute_industry_view(bullish_db)


@huey.task(context=True)  # type: ignore
@job_tracker
def backtest_signals(
    database_path: Path,
    job_type: JobType,
    task: Optional[Task] = None,
) -> None:
    bullish_db = BullishDb(database_path=database_path)
    run_signal_series_analysis(bullish_db)
    run_many_tests(bullish_db, predefined_filters(), BackTestConfig())


def base_news(
    database_path: Path,
    job_type: JobType,
    symbols: List[str],
    headless: bool = True,
    task: Optional[Task] = None,
) -> None:
    bullish_db = BullishDb(database_path=database_path)
    database_config = DatabaseConfig(database_path=database_path, no_migration=True)
    get_news(symbols, database_config, headless=headless, model_name="qwen3:4b")
    run_analysis(bullish_db)


@huey.task(context=True)  # type: ignore
@job_tracker
def news(
    database_path: Path,
    job_type: JobType,
    symbols: List[str],
    headless: bool = True,
    task: Optional[Task] = None,
) -> None:
    bullish_db = BullishDb(database_path=database_path)
    if get_open_ai_news(bullish_db, symbols):
        for symbol in symbols:
            subject = bullish_db.read_subject(symbol)
            if subject:
                logger.debug(
                    f"extracting news for {symbol} subject: {subject.model_dump()}"
                )
                try:
                    bullish_db.update_analysis(
                        symbol,
                        subject.model_dump(
                            exclude_none=True,
                            exclude_unset=True,
                            exclude_defaults=True,
                            exclude={"symbol"},
                        ),
                    )
                except Exception as e:
                    logger.error(f"failed to extract news for {symbol}: {e}")
                    print(f"failed to extract news for {symbol}: {e}")
                    continue


@huey.periodic_task(crontab(minute="0", hour="8"), context=True)  # type: ignore
def cron_news(
    task: Optional[Task] = None,
) -> None:
    filters = load_custom_filters()
    database = DataBaseSingleTon()
    if database.valid() and filters:
        bullish_db = BullishDb(database_path=database.path)
        data = pd.concat([bullish_db.read_filter_query(f) for f in filters])
        if not data.empty and "symbol" in data.columns:
            job_tracker(base_news)(
                database_path=database.path,
                job_type="Fetching news",
                symbols=data["symbol"].unique().tolist(),
                headless=False,
                task=task,
            )



"""

Revision ID: 12889a2cbd7d
Revises: 040b15fba458
Create Date: 2025-07-17 17:50:35.004785

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "12889a2cbd7d"
down_revision: Union[str, None] = "040b15fba458"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "signalseries",
        sa.Column("date", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("value", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("date", "name", "symbol"),
    )

    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_analysis_momentum_time_span"))
        batch_op.drop_column("momentum_time_span")

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("momentum_time_span", sa.DATE(), nullable=True))
        batch_op.create_index(
            batch_op.f("ix_analysis_momentum_time_span"),
            ["momentum_time_span"],
            unique=False,
        )

    op.drop_table("signalseries")
    # ### end Alembic commands ###



import datetime
import logging
from typing import (
    Optional,
    Any,
    Annotated,
    Literal,
    Dict,
    List,
    TYPE_CHECKING,
    get_args,
)

import numpy as np
import pandas as pd
from bearish.models.base import Ticker  # type: ignore
from bearish.models.price.prices import Prices  # type: ignore
from bearish.models.query.query import AssetQuery, Symbols  # type: ignore
from pydantic import BaseModel, BeforeValidator, Field, model_validator

from bullish.analysis.constants import (
    Industry,
    IndustryGroup,
    Sector,
    Country,
    WesternCountries,
)

if TYPE_CHECKING:
    from bullish.database.crud import BullishDb

logger = logging.getLogger(__name__)
Type = Literal["Mean"]

FUNCTIONS = {"Mean": np.mean}
BASELINE_DATE = datetime.date.today() - datetime.timedelta(days=60)


def compute_normalized_close(close_: pd.Series) -> pd.Series:
    close = close_.copy()
    close.index = close.index.tz_localize(None)  # type: ignore
    closest_ts = close.index[
        close.index.get_indexer([BASELINE_DATE], method="nearest")[0]
    ]
    normalized_close = (close / close.loc[closest_ts]).rename("normalized_close")
    normalized_close.index = close_.index
    return normalized_close  # type: ignore


def get_industry_comparison_data(
    bullish_db: "BullishDb",
    symbol_data: pd.DataFrame,
    type: Type,
    industry: Industry,
    country: Country,
) -> pd.DataFrame:
    try:

        views = bullish_db.read_returns(type, industry, country)
        industry_data = IndustryViews.from_views(views).to_dataframe()
        normalized_symbol = compute_normalized_close(symbol_data.close).rename("symbol")
        normalized_industry = industry_data.normalized_close.rename(industry)
        data = [normalized_symbol, normalized_industry]
        for country in get_args(WesternCountries):
            views = bullish_db.read_returns(type, industry, country)
            if views:
                industry_data = IndustryViews.from_views(views).to_dataframe()
                normalized_industry = industry_data.normalized_close.rename(
                    f"{industry}-{country}"
                )
                data.append(normalized_industry)
        return pd.concat(data, axis=1)
    except Exception as e:
        logger.error(e)
        return pd.DataFrame()


class PricesReturns(Prices):  # type: ignore

    def returns(self) -> pd.DataFrame:
        try:
            data = self.to_dataframe()
            data["simple_return"] = data.close.pct_change() * 100
            data["log_return"] = (data.close / data.close.shift(1)).apply(np.log) * 100
            data["normalized_close"] = compute_normalized_close(data.close)
            return data[["simple_return", "log_return", "normalized_close"]]  # type: ignore
        except Exception:
            return pd.DataFrame(
                columns=["simple_return", "log_return", "normalized_close"]
            )


def to_float(value: Any) -> Optional[float]:
    if value == "None":
        return None
    if value is None:
        return None
    if isinstance(value, str):
        try:
            return float(value)
        except ValueError:
            return None
    return float(value)


class Basedate(BaseModel):
    date: datetime.date


class KPI(BaseModel):
    simple_return: Annotated[float, BeforeValidator(to_float), Field(None)]
    log_return: Annotated[float, BeforeValidator(to_float), Field(None)]
    normalized_close: Annotated[float, BeforeValidator(to_float), Field(None)]


class BaseIndustryView(Basedate, KPI): ...


class IndustryView(BaseIndustryView):
    created_at: datetime.date
    country: Country
    industry: Industry
    industry_group: Optional[IndustryGroup] = None
    sector: Optional[Sector] = None
    type: Type

    @model_validator(mode="before")
    def _validate(cls, values: Dict[str, Any]) -> Dict[str, Any]:  # noqa: N805
        created_at = datetime.date.today()
        current_date = values.get("date", created_at)
        return (
            {"date": current_date}
            | values
            | {
                "created_at": created_at,
            }
        )

    @classmethod
    def from_data(
        cls,
        data: pd.DataFrame,
        function_name: Type,
        industry: Industry,
        country: Country,
    ) -> List["IndustryView"]:
        function = FUNCTIONS[function_name]
        data_ = []
        for field in KPI.model_fields:

            data__ = (
                data[field].apply(function, axis=1).rename(field)
                if data[[field]].shape[1] > 1
                else data[field]
            )

            data_.append(data__)

        data_final = pd.concat(data_, axis=1)
        data_final["date"] = data_final.index
        return [
            cls.model_validate(
                r | {"industry": industry, "type": function_name, "country": country}
            )
            for r in data_final.to_dict(orient="records")
        ]

    @classmethod
    def from_db(
        cls, bullish: "BullishDb", industry: Industry, country: Country
    ) -> List["IndustryView"]:
        returns = []
        symbols = bullish.read_industry_symbols(industries=[industry], country=country)
        query = AssetQuery(
            symbols=Symbols(equities=[Ticker(symbol=s) for s in symbols])
        )
        data = bullish.read_series(query, months=6)
        raw_data = [
            PricesReturns(prices=[d for d in data if d.symbol == s]).returns()
            for s in symbols
        ]
        raw_data = [r for r in raw_data if not r.empty]

        if raw_data:
            data_ = pd.concat(raw_data, axis=1)
            for function_name in FUNCTIONS:
                returns.extend(cls.from_data(data_, function_name, industry, country))  # type: ignore
        return returns


class IndustryViews(BaseModel):
    views: List[IndustryView]

    def to_dataframe(self) -> pd.DataFrame:
        data = pd.DataFrame.from_records(
            [
                p.model_dump(include=set(BaseIndustryView.model_fields))
                for p in self.views
            ]
        )
        if data.empty:
            return data
        data = data.set_index("date", inplace=False)
        data = data.sort_index(inplace=False)

        data.index = pd.to_datetime(data.index, utc=True)
        data = data[~data.index.duplicated(keep="first")]
        return data

    @classmethod
    def from_views(cls, views: List[IndustryView]) -> "IndustryViews":
        return cls(views=views)


def compute_industry_view(bullish: "BullishDb") -> None:
    for country in get_args(Country):
        for industry in get_args(Industry):
            returns = IndustryView.from_db(bullish, industry, country)
            if returns:
                bullish.write_returns(returns)



import os
from pathlib import Path

from alembic import command
from alembic.config import Config

from bullish.database.settings import TEST_DATABASE_URL


def create_revision(database_url: str, message: str) -> None:
    os.environ.update({"DATABASE_URL": database_url})
    root_folder = Path(__file__).parents[1]
    alembic_cfg = Config(root_folder / "alembic" / "alembic.ini")
    alembic_cfg.set_main_option("script_location", str(root_folder / "alembic"))
    command.revision(alembic_cfg, message=message, autogenerate=True)


if __name__ == "__main__":
    message = ""
    create_revision(TEST_DATABASE_URL, message)



"""

Revision ID: 4b0a2f40b7d3
Revises: 037dbd721317
Create Date: 2025-06-20 09:17:53.566652

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "4b0a2f40b7d3"
down_revision: Union[str, None] = "037dbd721317"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "jobtracker",
        sa.Column("job_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("type", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("status", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("started_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("job_id"),
    )
    op.drop_table("view")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("jobtracker")
    op.create_table(
        "view",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("isin", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("view_name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "source"),
    )
    # ### end Alembic commands ###



"""

Revision ID: d0e58e050845
Revises: ff0cc4ba40ec
Create Date: 2025-08-05 14:02:54.407561

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "d0e58e050845"
down_revision: Union[str, None] = "ff0cc4ba40ec"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("next_earnings_date", sa.Date(), nullable=True))
        batch_op.create_index(
            "ix_analysis_next_earnings_date", ["next_earnings_date"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_next_earnings_date")
        batch_op.drop_column("next_earnings_date")

    # ### end Alembic commands ###



"""

Revision ID: 040b15fba458
Revises: ec25c8fa449f
Create Date: 2025-07-15 09:22:11.456381

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "040b15fba458"
down_revision: Union[str, None] = "ec25c8fa449f"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "industryview",
        sa.Column("simple_return", sa.Float(), nullable=True),
        sa.Column("log_return", sa.Float(), nullable=True),
        sa.Column("normalized_close", sa.Float(), nullable=True),
        sa.Column("date", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("created_at", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("country", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("industry", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("industry_group", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("sector", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("type", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.PrimaryKeyConstraint("date", "country", "industry", "type"),
    )

    op.drop_table("industryreturns")


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    op.create_table(
        "industryreturns",
        sa.Column("date", sa.VARCHAR(), nullable=False),
        sa.Column("created_at", sa.VARCHAR(), nullable=True),
        sa.Column("simple_return", sa.FLOAT(), nullable=True),
        sa.Column("log_return", sa.FLOAT(), nullable=True),
        sa.Column("country", sa.VARCHAR(), nullable=False),
        sa.Column("industry", sa.VARCHAR(), nullable=False),
        sa.Column("industry_group", sa.VARCHAR(), nullable=True),
        sa.Column("sector", sa.VARCHAR(), nullable=True),
        sa.Column("type", sa.VARCHAR(), nullable=False),
        sa.PrimaryKeyConstraint("date", "country", "industry", "type"),
    )

    op.drop_table("industryview")
    # ### end Alembic commands ###



from typing import Literal

Industry = Literal[
    "Publishing",
    "Internet Retail",
    "Scientific & Technical Instruments",
    "Engineering & Construction",
    "Diagnostics & Research",
    "Software - Infrastructure",
    "Thermal Coal",
    "Software - Application",
    "Auto Manufacturers",
    "Farm Products",
    "Medical Devices",
    "Education & Training Services",
    "Auto Parts",
    "Specialty Chemicals",
    "Marine Shipping",
    "Biotechnology",
    "Real Estate Services",
    "Gold",
    "Entertainment",
    "Specialty Retail",
    "Utilities - Independent Power Producers",
    "Steel",
    "Mortgage Finance",
    "Communication Equipment",
    "Drug Manufacturers - Specialty & Generic",
    "Electronic Gaming & Multimedia",
    "Banks - Regional",
    "Oil & Gas E&P",
    "Travel Services",
    "Real Estate - Diversified",
    "Telecom Services",
    "Uranium",
    "Consulting Services",
    "Waste Management",
    "Agricultural Inputs",
    "Utilities - Diversified",
    "Auto & Truck Dealerships",
    "Confectioners",
    "Other Industrial Metals & Mining",
    "Beverages - Wineries & Distilleries",
    "Oil & Gas Midstream",
    "Recreational Vehicles",
    "Electrical Equipment & Parts",
    "Household & Personal Products",
    "Packaging & Containers",
    "REIT - Specialty",
    "Home Improvement Retail",
    "Electronic Components",
    "Asset Management",
    "Consumer Electronics",
    "Conglomerates",
    "Health Information Services",
    "Medical Instruments & Supplies",
    "Building Products & Equipment",
    "Information Technology Services",
    "Specialty Industrial Machinery",
    "Food Distribution",
    "Packaged Foods",
    "Rental & Leasing Services",
    "Medical Distribution",
    "Grocery Stores",
    "Advertising Agencies",
    "Beverages - Non - Alcoholic",
    "Apparel Manufacturing",
    "Oil & Gas Equipment & Services",
    "Coking Coal",
    "Industrial Distribution",
    "Restaurants",
    "Beverages - Brewers",
    "Chemicals",
    "Real Estate - Development",
    "Credit Services",
    "Tobacco",
    "Metal Fabrication",
    "Building Materials",
    "Residential Construction",
    "Specialty Business Services",
    "REIT - Hotel & Motel",
    "Internet Content & Information",
    "Lodging",
    "Furnishings, Fixtures & Appliances",
    "Airlines",
    "Computer Hardware",
    "Integrated Freight & Logistics",
    "Solar",
    "Capital Markets",
    "Leisure",
    "Airports & Air Services",
    "Aluminum",
    "Insurance Brokers",
    "Semiconductors",
    "REIT - Retail",
    "Luxury Goods",
    "Lumber & Wood Production",
    "REIT - Mortgage",
    "Semiconductor Equipment & Materials",
    "Aerospace & Defense",
    "Security & Protection Services",
    "Utilities - Renewable",
    "Utilities - Regulated Gas",
    "Apparel Retail",
    "Pollution & Treatment Controls",
    "Broadcasting",
    "Resorts & Casinos",
    "Other Precious Metals & Mining",
    "Financial Data & Stock Exchanges",
    "Footwear & Accessories",
    "Medical Care Facilities",
    "Electronics & Computer Distribution",
    "Gambling",
    "Tools & Accessories",
    "Insurance - Property & Casualty",
    "Utilities - Regulated Water",
    "Insurance - Specialty",
    "Personal Services",
    "Pharmaceutical Retailers",
    "Farm & Heavy Construction Machinery",
    "Utilities - Regulated Electric",
    "Department Stores",
    "Staffing & Employment Services",
    "Textile Manufacturing",
    "Silver",
    "REIT - Industrial",
    "REIT - Diversified",
    "Copper",
    "Business Equipment & Supplies",
    "Infrastructure Operations",
    "Trucking",
    "Insurance - Reinsurance",
    "Insurance - Diversified",
    "Drug Manufacturers - General",
    "Oil & Gas Drilling",
    "Banks - Diversified",
    "REIT - Residential",
    "Oil & Gas Refining & Marketing",
    "Shell Companies",
    "Financial Conglomerates",
    "Paper & Paper Products",
    "Insurance - Life",
    "REIT - Office",
    "Railroads",
    "Oil & Gas Integrated",
    "Healthcare Plans",
    "REIT - Healthcare Facilities",
    "Discount Stores",
]
IndustryGroup = Literal[
    "publishing",
    "internet-retail",
    "scientific-technical-instruments",
    "engineering-construction",
    "diagnostics-research",
    "software-infrastructure",
    "thermal-coal",
    "software-application",
    "auto-manufacturers",
    "farm-products",
    "medical-devices",
    "education-training-services",
    "auto-parts",
    "specialty-chemicals",
    "marine-shipping",
    "biotechnology",
    "real-estate-services",
    "gold",
    "entertainment",
    "specialty-retail",
    "utilities-independent-power-producers",
    "steel",
    "mortgage-finance",
    "communication-equipment",
    "drug-manufacturers-specialty-generic",
    "electronic-gaming-multimedia",
    "banks-regional",
    "oil-gas-e-p",
    "travel-services",
    "real-estate-diversified",
    "telecom-services",
    "uranium",
    "consulting-services",
    "waste-management",
    "agricultural-inputs",
    "utilities-diversified",
    "auto-truck-dealerships",
    "confectioners",
    "other-industrial-metals-mining",
    "beverages-wineries-distilleries",
    "oil-gas-midstream",
    "recreational-vehicles",
    "electrical-equipment-parts",
    "household-personal-products",
    "packaging-containers",
    "reit-specialty",
    "home-improvement-retail",
    "electronic-components",
    "asset-management",
    "consumer-electronics",
    "conglomerates",
    "health-information-services",
    "medical-instruments-supplies",
    "building-products-equipment",
    "information-technology-services",
    "specialty-industrial-machinery",
    "food-distribution",
    "packaged-foods",
    "rental-leasing-services",
    "medical-distribution",
    "grocery-stores",
    "advertising-agencies",
    "beverages-non-alcoholic",
    "apparel-manufacturing",
    "oil-gas-equipment-services",
    "coking-coal",
    "industrial-distribution",
    "restaurants",
    "beverages-brewers",
    "chemicals",
    "real-estate-development",
    "credit-services",
    "tobacco",
    "metal-fabrication",
    "building-materials",
    "residential-construction",
    "specialty-business-services",
    "reit-hotel-motel",
    "internet-content-information",
    "lodging",
    "furnishings-fixtures-appliances",
    "airlines",
    "computer-hardware",
    "integrated-freight-logistics",
    "solar",
    "capital-markets",
    "leisure",
    "airports-air-services",
    "aluminum",
    "insurance-brokers",
    "semiconductors",
    "reit-retail",
    "luxury-goods",
    "lumber-wood-production",
    "reit-mortgage",
    "semiconductor-equipment-materials",
    "aerospace-defense",
    "security-protection-services",
    "utilities-renewable",
    "utilities-regulated-gas",
    "apparel-retail",
    "pollution-treatment-controls",
    "broadcasting",
    "resorts-casinos",
    "other-precious-metals-mining",
    "financial-data-stock-exchanges",
    "footwear-accessories",
    "medical-care-facilities",
    "electronics-computer-distribution",
    "gambling",
    "tools-accessories",
    "insurance-property-casualty",
    "utilities-regulated-water",
    "insurance-specialty",
    "personal-services",
    "pharmaceutical-retailers",
    "farm-heavy-construction-machinery",
    "utilities-regulated-electric",
    "department-stores",
    "staffing-employment-services",
    "textile-manufacturing",
    "silver",
    "reit-industrial",
    "reit-diversified",
    "copper",
    "business-equipment-supplies",
    "infrastructure-operations",
    "trucking",
    "insurance-reinsurance",
    "insurance-diversified",
    "drug-manufacturers-general",
    "oil-gas-drilling",
    "banks-diversified",
    "reit-residential",
    "oil-gas-refining-marketing",
    "shell-companies",
    "financial-conglomerates",
    "paper-paper-products",
    "insurance-life",
    "reit-office",
    "railroads",
    "oil-gas-integrated",
    "healthcare-plans",
    "reit-healthcare-facilities",
    "discount-stores",
]
Sector = Literal[
    "Communication Services",
    "Consumer Cyclical",
    "Technology",
    "Industrials",
    "Healthcare",
    "Energy",
    "Consumer Defensive",
    "Basic Materials",
    "Real Estate",
    "Utilities",
    "Financial Services",
    "Conglomerates",
]
Europe = Literal["Germany", "Belgium", "France"]
Us = Literal["United states"]
WesternCountries = Literal["United kingdom", Europe, Us]
Country = Literal[
    WesternCountries,
    "Australia",
    "China",
    "Japan",
    "Poland",
    "Switzerland",
    "Canada",
    "Greece",
    "Spain",
    "Indonesia",
    "Netherlands",
    "British virgin islands",
    "Italy",
    "Hungary",
    "Austria",
    "Finland",
    "Sweden",
    "Bermuda",
    "Taiwan",
    "Israel",
    "Ukraine",
    "Singapore",
    "Jersey",
    "Ireland",
    "Luxembourg",
    "Cyprus",
    "Cayman islands",
    "Norway",
    "Denmark",
    "Hong kong",
    "New zealand",
    "Kazakhstan",
    "Nigeria",
    "Argentina",
    "Brazil",
    "Czech republic",
    "Mauritius",
    "South africa",
    "India",
    "Mexico",
    "Mongolia",
    "Slovenia",
    "Thailand",
    "Malaysia",
    "Costa rica",
    "Isle of man",
    "Egypt",
    "Turkey",
    "United arab emirates",
    "Colombia",
    "Gibraltar",
    "Malta",
    "Liechtenstein",
    "Guernsey",
    "Peru",
    "Estonia",
    "French guiana",
    "Portugal",
    "Uruguay",
    "Chile",
    "Martinique",
    "Monaco",
    "Panama",
    "Papua new guinea",
    "South korea",
    "Macau",
    "Gabon",
    "Romania",
    "Senegal",
    "Morocco",
    "Jordan",
    "Lithuania",
    "Dominican republic",
    "Reunion",
    "Zambia",
    "Cambodia",
    "Myanmar",
    "Bahamas",
    "Philippines",
    "Bangladesh",
    "Latvia",
    "Vietnam",
    "Iceland",
    "Azerbaijan",
    "Georgia",
    "Liberia",
    "Kenya",
]


HighGrowthIndustry = Literal[
    "Software - Infrastructure",
    "Software - Application",
    "Internet Retail",
    "Internet Content & Information",
    "Electronic Gaming & Multimedia",
    "Semiconductors",
    "Semiconductor Equipment & Materials",
    "Information Technology Services",
    "Communication Equipment",
    "Consumer Electronics",
    "Health Information Services",
    "Biotechnology",
    "Medical Devices",
    "Diagnostics & Research",
    "Medical Instruments & Supplies",
    "Drug Manufacturers - Specialty & Generic",
]

DefensiveIndustries = Literal[
    "Utilities - Independent Power Producers",
    "Utilities - Diversified",
    "Utilities - Renewable",
    "Utilities - Regulated Gas",
    "Utilities - Regulated Water",
    "Utilities - Regulated Electric",
    "Household & Personal Products",
    "Food Distribution",
    "Packaged Foods",
    "Grocery Stores",
    "Beverages - Non - Alcoholic",
    "Beverages - Brewers",
    "Confectioners",
    "Tobacco",
    "Paper & Paper Products",
    "Medical Devices",
    "Drug Manufacturers - Specialty & Generic",
    "Medical Instruments & Supplies",
    "Medical Distribution",
    "Medical Care Facilities",
    "Drug Manufacturers - General",
    "Healthcare Plans",
    "Pharmaceutical Retailers",
    "Waste Management",
    "Pollution & Treatment Controls",
    "Insurance Brokers",
    "Insurance - Property & Casualty",
    "Insurance - Specialty",
    "Insurance - Reinsurance",
    "Insurance - Diversified",
    "Insurance - Life",
]



from datetime import datetime
from typing import Literal, get_args

import pandas as pd
from pydantic import BaseModel, Field

JobType = Literal[
    "Update data", "Update analysis", "Fetching news", "backtest signals", "Initialize"
]
JobStatus = Literal["Completed", "Failed", "Running", "Started"]
StatusIcon = ["✅ Completed", "❌ Failed", "🔄 Running", "🚀 Started"]


class JobTrackerStatus(BaseModel):
    job_id: str
    status: JobStatus = "Started"


class JobTracker(JobTrackerStatus):
    type: JobType
    started_at: datetime = Field(default_factory=datetime.now)


def add_icons(data: pd.DataFrame) -> pd.DataFrame:
    status_map = dict(zip(list(get_args(JobStatus)), StatusIcon, strict=True))
    data["status"] = data["status"].map(status_map)
    return data
